from collections import defaultdict
import copy

import numpy as np

from typing import Optional
import gym
from gym import core
from gym.utils import seeding

from elastica._calculus import _isnan_check
from elastica.timestepper import extend_stepper_interface
from elastica import *

from gym_softrobot import RENDERER_CONFIG
from gym_softrobot.config import RendererType
from gym_softrobot.utils.render.base_renderer import (
    BaseRenderer,
    BaseElasticaRendererSession,
)

from gym_softrobot.utils.custom_elastica.muscle_torque import (
    MuscleTorquesWithVaryingBetaSplines,
)


# Set base simulator class
class BaseSimulator(BaseSystemCollection, Constraints, Connections, Forcing, CallBacks):
    pass


# TODO: generalize this as a Class for online trajectory generation.
# TODO: remove np.random: use fixed generator from seed value. (determinism)
def generate_trajectory(final_time, sim_dt, target_v_scale):
    end_time = (
        final_time * 1.1
    )  # Use of 1.1 just adds a little buffer on the end of the trajectory.
    numpoints = np.rint(1 / sim_dt * end_time).astype(int)
    target_trajectory = np.zeros((numpoints, 3))

    """ Generate a trajectory bounded in a box [-0.8, 0.8] x [-0.8, 0.8] x [0.0, 0.8] """
    t = np.array([end_time * float(i) / (numpoints - 1) for i in range(numpoints)])
    t += (
        np.random.rand(1)[0] * 3600
    )  # start the trajectory anywhere within a 1 hour period

    f1 = np.random.uniform(2, 5) * 0.025 * target_v_scale
    f2 = np.random.uniform(2, 5) * 0.025 * target_v_scale
    f3 = np.random.uniform(2, 5) * 0.025 * target_v_scale
    direction = np.random.randint(0, 1) * 2 - 1
    target_trajectory[:, 0] = (
        direction
        * 0.8
        * np.sin(2 * np.pi * f1 * t)
        * np.sin(2 * np.pi * f2 * t)
        * np.sin(2 * np.pi * f3 * t)
        * 1000
    )

    f1 = np.random.uniform(2, 5) * 0.025 * target_v_scale
    f2 = np.random.uniform(2, 5) * 0.025 * target_v_scale
    f3 = np.random.uniform(2, 5) * 0.025 * target_v_scale
    direction = np.random.randint(0, 1) * 2 - 1
    target_trajectory[:, 1] = (
        direction
        * 0.4
        * np.sin(2 * np.pi * f1 * t)
        * np.sin(2 * np.pi * f2 * t)
        * np.sin(2 * np.pi * f3 * t)
        * 1000
    )
    target_trajectory[:, 1] += 0.4

    f1 = np.random.uniform(2, 5) * 0.025 * target_v_scale
    f2 = np.random.uniform(2, 5) * 0.025 * target_v_scale
    f3 = np.random.uniform(2, 5) * 0.025 * target_v_scale
    direction = np.random.randint(0, 1) * 2 - 1
    target_trajectory[:, 2] = (
        direction
        * 0.8
        * np.sin(2 * np.pi * f1 * t)
        * np.sin(2 * np.pi * f2 * t)
        * np.sin(2 * np.pi * f3 * t)
        * 1000
    )

    return target_trajectory


class SoftArmTrackingEnv(core.Env):
    metadata = {"render.modes": ["rgb_array", "human"]}

    def __init__(self, game_mode: int = 1):
        self.n_elem = 40
        self.sim_dt = 2.0e-4
        self.RL_update_interval = 0.01  # This is 100 updates per second
        self.num_steps_per_update = np.rint(
            self.RL_update_interval / self.sim_dt
        ).astype(int)
        self.youngs_modulus = 2e6
        poisson_ratio = 0.45

        self.torque_scale = 10

        self.max_episode_final_time = 5  # seconds

        self.base_length = 1000
        self.radius = 50
        self.COLLECT_DATA_FOR_POSTPROCESSING = False

        self.number_of_control_points = 4
        self.number_of_observation_segments = self.number_of_control_points

        self.rendering_fps = 30
        self.step_skip = np.rint(1.0 / (self.rendering_fps * self.sim_dt)).astype(int)

        self.max_rate_of_change_of_activation = np.infty
        self.target_v_scale = 0.1

        self.mode = game_mode
        self.target_location = np.array([500, 500.0, 500])

        self.StatefulStepper = PositionVerlet()

        # normal and/or binormal direction activation (3D)
        self.action_space = gym.spaces.Box(
            low=-1.0,
            high=1.0,
            shape=(2 * self.number_of_control_points,),
            dtype=np.float64,
        )
        self.observation_space = gym.spaces.Box(
            low=-np.inf,
            high=np.inf,
            shape=(self.number_of_observation_segments * 2 + 6,),
            dtype=np.float64,
        )

        # Determinism
        self.seed()

        # Rendering-related
        self.viewer = None
        self.renderer = None

    def seed(self, seed=None):
        # Deprecated in new gym
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def get_state(self):
        """
        Returns current state of the system to the controller.

        Returns
        -------
        numpy.ndarray
            1D (number_of_states) array containing data with 'float' type.
            Size of the states depends on the problem.
        """
        avg_kappa_1 = np.zeros(self.number_of_observation_segments)
        # self.n_elem-1 is number of curvature values due to veroni regions
        avg_length = int((self.n_elem - 1) / (self.number_of_observation_segments))

        for i in range(0, self.number_of_observation_segments - 1):
            lower = np.rint(avg_length * (i)).astype(int)
            upper = np.rint(avg_length * (i + 1)).astype(int)
            avg_kappa_1[i] = self.shearable_rod.kappa[0, lower:upper].mean()
        lower = np.rint(avg_length * (self.number_of_observation_segments - 1)).astype(
            int
        )
        avg_kappa_1[-1] = self.shearable_rod.kappa[0, lower:].mean()

        avg_kappa_2 = np.zeros(self.number_of_observation_segments)
        for i in range(0, self.number_of_observation_segments - 1):
            lower = np.rint(avg_length * (i)).astype(int)
            upper = np.rint(avg_length * (i + 1)).astype(int)
            avg_kappa_2[i] = self.shearable_rod.kappa[1, lower:upper].mean()
        lower = np.rint(avg_length * (self.number_of_observation_segments - 1)).astype(
            int
        )
        avg_kappa_2[-1] = self.shearable_rod.kappa[1, lower:].mean()

        state = np.concatenate(
            (
                # rod curvature information
                avg_kappa_1
                * self.base_length
                / (
                    2 * np.pi
                ),  # normalize by the curvature of if the rod is a perfect ring.
                avg_kappa_2
                * self.base_length
                / (
                    2 * np.pi
                ),  # normalize by the curvature of if the rod is a perfect ring.
                # arm tip location
                self.shearable_rod.position_collection[..., -1] / self.base_length,
                # target location
                self.wsol[self.tick] / 1000,  # convert back to meters (~[-1,1])
            )
        )

        return state

    def step(self, action):
        # set binormal activations to 0 if solving 2D case
        self.spline_points_func_array_normal_dir[:] = action[
            : self.number_of_control_points
        ]
        self.spline_points_func_array_binormal_dir[:] = action[
            self.number_of_control_points :
        ]

        for _ in range(int(self.num_steps_per_update)):
            # print('step:', i)
            self.time_tracker = self.do_step(
                self.StatefulStepper,
                self.stages_and_updates,
                self.simulator,
                self.time_tracker,
                self.sim_dt,
            )
            self.tick += 1
            self.sphere.position_collection[..., 0] = self.wsol[self.tick]
            self.sphere.velocity_collection[..., 0] = self.velocity_sphere[self.tick]

        """ Reward Engineering """
        tip_to_target = (
            self.sphere.position_collection[..., 0]
            - self.shearable_rod.position_collection[..., -1]
        ) / 1000
        reward_dist = -np.square(np.linalg.norm(tip_to_target))
        reward = 1.0 * reward_dist  # + reward_orientation

        # observe current state: current as sensed signal
        state = self.get_state()

        """ Done is a boolean to reset the environment before episode is completed """
        terminated = False
        truncated = False

        # Position of the rod cannot be NaN, it is not valid, stop the simulation
        invalid_values_condition = _isnan_check(state)
        if invalid_values_condition == True:
            reward = -100
            state = np.nan_to_num(self.get_state())
            terminated = True
            truncated = True
            print("Episode blew up. Maybe try a smaller dt?")

        if self.tick * self.sim_dt >= self.max_episode_final_time:
            terminated = True
            truncated = True
            print("Episode has reached max time")

        self._target = self.wsol[self.tick]
        return state, reward, terminated, truncated, {"ctime": self.time_tracker}

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        return_info: bool = False,
        options: Optional[dict] = None,
    ):
        super().reset(seed=seed)
        self.simulator = BaseSimulator()

        ###--------------ADD ARM TO SIMULATION--------------###
        start = np.zeros((3,))
        direction = np.array([0.0, 1.0, 0.0])  # rod direction: pointing upwards
        normal = np.array([0.0, 0.0, 1.0])
        binormal = np.cross(direction, normal)

        # Set the arm properties after defining rods

        damping = self.youngs_modulus * 1e-7 * 1

        radius_tip = self.radius  # radius of the arm at the tip
        radius_base = radius_tip  # radius of the arm at the base
        radius_along_rod = np.linspace(radius_tip, radius_tip, self.n_elem)

        # Arm is shearable Cosserat rod
        self.shearable_rod = CosseratRod.straight_rod(
            self.n_elem,
            start,
            direction,
            normal,
            self.base_length,
            base_radius=radius_along_rod,
            density=1000 * 1e-6,
            # nu=damping,
            youngs_modulus=self.youngs_modulus,
        )

        self.shearable_rod.dissipation_constant_for_torques *= (
            1e6  # accounts for the new g/mm/s units (compared to kg/m/s)
        )
        self.simulator.append(
            self.shearable_rod
        )  # Now rod is ready for simulation, append rod to simulation
        self.simulator.constrain(self.shearable_rod).using(
            OneEndFixedRod, constrained_position_idx=(0,), constrained_director_idx=(0,)
        )
        # self.simulator.add_forcing_to(self.shearable_rod).using(GravityForces, acc_gravity=np.array([0.0, 9.81*1e3, 0.0]))

        # Call back function to collect arm data from simulation
        if self.COLLECT_DATA_FOR_POSTPROCESSING:

            class ArmCallBack(CallBackBaseClass):
                """
                Call back function for Elastica rod
                """

                def __init__(
                    self,
                    step_skip: int,
                    callback_params: dict,
                ):
                    CallBackBaseClass.__init__(self)
                    self.every = step_skip
                    self.callback_params = callback_params

                def make_callback(self, system, time, current_step: int):
                    if current_step % self.every == 0:
                        self.callback_params["time"].append(time)
                        self.callback_params["step"].append(current_step)
                        self.callback_params["position"].append(
                            system.position_collection.copy()
                        )
                        self.callback_params["radius"].append(system.radius.copy())
                        self.callback_params["com"].append(
                            system.compute_position_center_of_mass()
                        )
                        self.callback_params["directors"].append(
                            system.director_collection.copy()
                        )
                        self.callback_params["kappa"].append(system.kappa.copy())
                        self.callback_params["omega_collection"].append(
                            system.omega_collection.copy()
                        )
                        self.callback_params["sigma"].append(system.sigma.copy())
                        self.callback_params["tangents"].append(system.tangents.copy())
                        self.callback_params["velocity_collection"].append(
                            system.velocity_collection.copy()
                        )
                        self.callback_params["acceleration_collection"].append(
                            system.acceleration_collection.copy()
                        )
                        return

            # Collect data using callback function for postprocessing
            self.post_processing_dict_rod = defaultdict(list)
            self.simulator.collect_diagnostics(self.shearable_rod).using(
                ArmCallBack,
                step_skip=self.step_skip,
                callback_params=self.post_processing_dict_rod,
            )

        """
        heuristic scaling of torque - not this works for a 25:1 slenderness ratio. Bending stiffness scales ~r^4 so if thicker or thinner,
        the torque value might need to be adjusted. This scaling is more designed for changing the youngs modulus
        """
        self.alpha = self.torque_scale * self.radius * self.youngs_modulus
        # Add muscle torques acting on the arm for actuation
        # MuscleTorquesWithVaryingBetaSplines uses the control points selected by RL to
        # generate torques along the arm.
        self.torque_profile_list_for_muscle_in_normal_dir = defaultdict(list)
        self.spline_points_func_array_normal_dir = []
        self.simulator.add_forcing_to(self.shearable_rod).using(
            MuscleTorquesWithVaryingBetaSplines,
            base_length=self.base_length,
            number_of_control_points=self.number_of_control_points,
            points_func_array=self.spline_points_func_array_normal_dir,
            muscle_torque_scale=self.alpha,
            direction=str("normal"),
            step_skip=self.step_skip,
            max_rate_of_change_of_activation=self.max_rate_of_change_of_activation,
            # max_signal_rate_of_change=4*self.sim_dt,
            torque_profile_recorder=self.torque_profile_list_for_muscle_in_normal_dir,
        )

        self.torque_profile_list_for_muscle_in_binormal_dir = defaultdict(list)
        self.spline_points_func_array_binormal_dir = []
        self.simulator.add_forcing_to(self.shearable_rod).using(
            MuscleTorquesWithVaryingBetaSplines,
            base_length=self.base_length,
            number_of_control_points=self.number_of_control_points,
            points_func_array=self.spline_points_func_array_binormal_dir,
            muscle_torque_scale=self.alpha,
            direction=str("binormal"),
            step_skip=self.step_skip,
            max_rate_of_change_of_activation=self.max_rate_of_change_of_activation,
            # max_signal_rate_of_change=4*self.sim_dt,
            torque_profile_recorder=self.torque_profile_list_for_muscle_in_binormal_dir,
        )

        ###--------------GENERATE TARGET TRAJECTORY--------------###
        self.tick = 0

        if self.mode == 2:
            # TODO: change this to a object that will update thee target position as you go
            target_trajectory = generate_trajectory(
                self.max_episode_final_time, self.sim_dt, self.target_v_scale
            )
            self.wsol = target_trajectory
        elif self.mode == 1:
            end_time = self.max_episode_final_time * 1.1
            numpoints = np.rint(1 / self.sim_dt * end_time).astype(int)
            self.wsol = np.zeros((numpoints, 3))
            self.wsol[:] = self.target_location

        velocity_sphere = self.wsol * 0
        velocity_sphere[:-1] = (self.wsol[1:] - self.wsol[:-1]) / self.sim_dt
        velocity_sphere[-1] = velocity_sphere[-2]
        self.velocity_sphere = velocity_sphere

        ###--------------ADD TARGET TO SIMULATION--------------###
        self.sphere_radius = 0.025 * 1000
        target_position = self.wsol[0]
        # initialize sphere
        self.sphere = Sphere(
            center=target_position,  # initialize target position of the ball
            base_radius=self.sphere_radius,
            density=1000 * 1e-6,
        )
        self.simulator.append(self.sphere)
        self.sphere.velocity_collection[..., 0] = self.velocity_sphere[0]

        if self.COLLECT_DATA_FOR_POSTPROCESSING:
            # Call back function to collect target sphere data from simulation
            class RigidSphereCallBack(CallBackBaseClass):
                """
                Call back function for target sphere
                """

                def __init__(self, step_skip: int, callback_params: dict):
                    CallBackBaseClass.__init__(self)
                    self.every = step_skip
                    self.callback_params = callback_params

                def make_callback(self, system, time, current_step: int):
                    if current_step % self.every == 0:
                        self.callback_params["time"].append(time)
                        self.callback_params["step"].append(current_step)
                        self.callback_params["position"].append(
                            system.position_collection.copy()
                        )
                        self.callback_params["velocity"].append(
                            system.velocity_collection.copy()
                        )
                        self.callback_params["radius"].append(
                            copy.deepcopy(system.radius)
                        )
                        self.callback_params["com"].append(
                            system.compute_position_center_of_mass()
                        )
                        return

            self.post_processing_dict_sphere = defaultdict(list)
            self.simulator.collect_diagnostics(self.sphere).using(
                RigidSphereCallBack,
                step_skip=self.step_skip,
                callback_params=self.post_processing_dict_sphere,
            )

        ###--------------FINALIZE SIMULATION--------------###
        # Finalize simulation environment. After finalize, you cannot add
        # any forcing, constrain or call back functions
        self.simulator.finalize()

        # do_step, stages_and_updates will be used in step function
        self.do_step, self.stages_and_updates = extend_stepper_interface(
            self.StatefulStepper, self.simulator
        )
        self.time_tracker = np.float64(0.0)

        state = self.get_state()
        self._target = self.wsol[self.tick]

        if return_info:
            return state, {}
        else:
            return state

    def render(self, mode="human", close=False):
        maxwidth = 800
        aspect_ratio = 3 / 4

        if self.viewer is None:
            from gym_softrobot.utils.render import pyglet_rendering

            self.viewer = pyglet_rendering.SimpleImageViewer(maxwidth=maxwidth)

        if self.renderer is None:
            # Switch renderer depending on configuration
            if RENDERER_CONFIG == RendererType.POVRAY:
                from gym_softrobot.utils.render.povray_renderer import Session
            elif RENDERER_CONFIG == RendererType.MATPLOTLIB:
                from gym_softrobot.utils.render.matplotlib_renderer import Session
            else:
                raise NotImplementedError("Rendering module is not imported properly")
            assert issubclass(Session, BaseRenderer), (
                "Rendering module is not properly subclassed"
            )
            assert issubclass(Session, BaseElasticaRendererSession), (
                "Rendering module is not properly subclassed"
            )
            self.renderer = Session(width=maxwidth, height=int(maxwidth * aspect_ratio))
            self.renderer.add_rod(self.shearable_rod)
            self.renderer.add_point(self._target.tolist(), self.sphere_radius)

        # POVRAY
        if RENDERER_CONFIG == RendererType.POVRAY:
            state_image = self.renderer.render(
                maxwidth,
                int(maxwidth * aspect_ratio * 0.7),
                camera_param=(
                    "location",
                    [50.0, 10.0, -60.0],
                    "look_at",
                    [-3.5, 5.7, 10.0],
                    "angle",
                    30,
                ),
            )
        elif RENDERER_CONFIG == RendererType.MATPLOTLIB:
            state_image = self.renderer.render()
        else:
            raise NotImplementedError("Rendering module is not imported properly")

        self.viewer.imshow(state_image)

        return state_image

    def close(self):
        if self.viewer:
            self.viewer.close()
            self.viewer = None
        if self.renderer:
            self.renderer.close()
            self.renderer = None



"""
Created on Jan. 07, 2021
@author: Heng-Sheng (Hanson) Chang
"""

import numpy as np
from numba import njit
import multiprocessing as mp

from elastica._linalg import _batch_cross
from elastica._calculus import quadrature_kernel
from elastica.external_forces import inplace_addition, NoForces

from gym_softrobot.utils.actuation.actuations.actuation import ContinuousActuation, ApplyActuation

from gym_softrobot.utils.actuation.frame_tools import change_box_to_arrow_axes

class Muscle(ContinuousActuation):
    def __init__(self, n_elements):
        ContinuousActuation.__init__(self, n_elements)
        self.n_elements = n_elements
        self.activation = np.zeros(self.n_elements-1)
        self.s = np.linspace(0, 1, self.n_elements+1)[1:-1]

    def set_activation(self, activation):
        self.activation[:] = activation
        return
    
    def get_activation(self):
        raise NotImplementedError

class MuscleForce(Muscle):
    def __init__(self, n_elements):
        Muscle.__init__(self, n_elements)
        self.distributed_activation = np.zeros(self.n_elements)

    def get_activation(self):
        redistribute_activation(self.activation, self.distributed_activation)
        return self.distributed_activation

@njit(cache=True)
def redistribute_activation(activation, distributed_activation):
    distributed_activation[0] = activation[0]/2
    distributed_activation[-1] = activation[-1]/2
    distributed_activation[1:-1] = (activation[1:] + activation[:-1])/2
    return

class MuscleCouple(Muscle):
    def __init__(self, n_elements):
        Muscle.__init__(self, n_elements)
    
    def get_activation(self):
        return self.activation

class MuscleFibers(object):
    def __init__(self, n_elements, control_numbers):
        self.controls = np.zeros(control_numbers)
        # self.G_activations = np.zeros((control_numbers, n_elements))
        self.G_internal_forces = np.zeros((control_numbers, 3, n_elements))
        self.G_internal_couples = np.zeros((control_numbers, 3, n_elements-1))
        self.G_external_forces = np.zeros((control_numbers, 3, n_elements+1))
        self.G_external_couples = np.zeros((control_numbers, 3, n_elements))
        self.G_flag = False

    @staticmethod
    @njit(cache=True)
    def muscle_fibers_function(
        controls,
        G_internal_forces, G_internal_couples,
        G_external_forces, G_external_couples,
        internal_forces, internal_couples,
        external_forces, external_couples
    ):
        for index in range(controls.shape[0]):
            inplace_addition(internal_forces, controls[index] * G_internal_forces[index, :, :])
            inplace_addition(internal_couples, controls[index] * G_internal_couples[index, :, :])
            inplace_addition(external_forces, controls[index] * G_external_forces[index, :, :])
            inplace_addition(external_couples, controls[index] * G_external_couples[index, :, :])

class ApplyMuscle(ApplyActuation):
    def __init__(self, muscles, step_skip: int, callback_params_list: list):
        ApplyActuation.__init__(self, None, step_skip, None)
        self.muscles = muscles
        self.callback_params_list = callback_params_list


    def apply_torques(self, system, time: float = 0.0):
        for muscle in self.muscles:
            muscle(system)
            inplace_addition(system.external_forces, muscle.external_forces)
            inplace_addition(system.external_torques, muscle.external_couples)

        self.make_callback()

    def callback_func(self):
        for muscle, callback_params in zip(self.muscles, self.callback_params_list):
            callback_params['activation'].append(muscle.activation.copy())
            callback_params['internal_force'].append(muscle.internal_forces.copy())
            callback_params['internal_couple'].append(muscle.internal_couples.copy())
            callback_params['external_force'].append(muscle.external_forces.copy())
            callback_params['external_couple'].append(muscle.external_couples.copy())

    def apply_muscles(self, muscle, system):
        muscle(system)
        inplace_addition(system.external_forces, muscle.external_forces)
        inplace_addition(system.external_torques, muscle.external_couples)

@njit(cache=True)
def local_strain(off_center_displacement, strain, curvature):
    return strain + _batch_cross(
        quadrature_kernel(curvature),
        off_center_displacement)

@njit(cache=True)
def force_length_curve_guassian(strain, intrinsic_strain=1, sigma=0.25):
    # blocksize = strain.shape[0]
    # force_weight = np.zeros(blocksize)
    # for i in range(blocksize):
    #     if strain[i] < intrinsic_strain:
    #         force_weight[i] = np.exp(-0.5*((strain[i]-intrinsic_strain)/sigma)**2)
    #     else:
    #         force_weight[i] = np.exp(-0.5*((strain[i]-intrinsic_strain)/(4*sigma))**2)
    force_weight = np.exp(-0.5*((strain-intrinsic_strain)/sigma)**2)
    return force_weight

@njit(cache=True)
def force_velocity_curve_tanh(strain_rate, max_strain_rate=1):
    force_weight = 1 + np.tanh(strain_rate/max_strain_rate)
    return force_weight

@njit(cache=True)
def passive_force_curve(strain, intrinsic_strain=1, coefficient=1, power=2):
    blocksize = strain.shape[0]
    force_weight = np.zeros(blocksize)
    for i in range(blocksize):
        if strain[i] >= intrinsic_strain:
            force_weight[i] = coefficient*(strain[i]-intrinsic_strain)**power
    return force_weight

# force-length curve (x) = 3.06 x^3 - 13.64 x^2 + 18.01 x - 6.44
@njit(cache=True)
def force_length_curve_poly(stretch, intrinsic_stretch=1, f_l_coefficients=np.array([-6.44, 18.01, -13.64, 3.06])):
    degree = f_l_coefficients.shape[0]

    blocksize = stretch.shape[0]
    force_weight = np.zeros(blocksize)
    for i in range(blocksize):
        for power in range(degree):
            force_weight[i] += f_l_coefficients[power] * (stretch[i] ** power)
        force_weight[i] = 0 if force_weight[i] < 0 else force_weight[i]
    return force_weight

@njit(cache=True)
def force_length_curve_hill(stretch, intrinsic_stretch=1):
    strain_rate_min = -42  # 1/sec
    c3 = 1450e3
    c4 = -625e3
    eps_c = 0.773
    c2 = c3 * eps_c / (c3 * eps_c + c4)
    c1 = c3 / (c2 * eps_c ** (c2 - 1))

    l_bz = 0.14e-6
    l_z = 0.06e-6
    # l0_sarc = l_act + l_z + 0.5*l_bz
    l0_sarc = 1.0e-6
    l_act = l0_sarc - l_z - 0.5 * l_bz

    l_myo = 0.7e-6
    D_act = 0.68
    D_myo = 1.90
    C_myo = 0.44
    l_max = l_myo + l_act + l_z
    l_min = l_bz
    f_a = 1
    k = 0.25
    E_comp = 2e4

    blocksize = stretch.shape[0]
    force_weight = np.zeros(blocksize)

    for i in range(blocksize):
        eps_r = stretch[i] - intrinsic_stretch
        l_sarc = l0_sarc + eps_r * l0_sarc
        #     print(l_sarc)
        if l_max < l_sarc:
            f_l = 0
        if l_act + l_bz + l_z <= l_sarc and l_sarc <= l_myo + l_act + l_z:
            f_l = (l_myo + l_act + l_z - l_sarc) / (l_myo - l_bz)
        if l_act + l_z <= l_sarc and l_sarc <= l_act + l_bz + l_z:
            f_l = 1
        if l_myo + l_z <= l_sarc and l_sarc <= l_act + l_z:
            f_l = (l_myo - l_bz - D_act * (l_act + l_z - l_sarc)) / (l_myo - l_bz)
        if l_min <= l_sarc and l_sarc <= l_myo + l_z:
            f_l = (
                l_myo
                - l_bz
                - D_act * (l_act + l_z - l_sarc)
                - D_myo * (l_myo + l_z - l_sarc)
            ) / (l_myo - l_bz) - C_myo * (l_myo + l_z - l_sarc) / (l_myo - l_bz)
        if l_sarc <= l_min:
            f_l = 0
        if f_l < 0:
            f_l = 0

        force_weight[i] = f_l

    return force_weight

def plot_force_length_active_curve():
    strain_range = np.linspace(0, 2, 1000)
    force_length_weight = force_length_curve_guassian(strain_range)
    passive_force_weight = passive_force_curve(strain_range)
    
    figsize = (19.2, 10.8)
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    ax.plot(strain_range, force_length_weight, color='C1', linewidth=8, label='active')
    ax.legend(prop={"size":40}, frameon=False)
    ax.set_xlim(-0.1, 2.1)
    ax.tick_params(axis='x', labelsize=48)
    ax.tick_params(axis='y', labelsize=48)
    ax.set_yticks([0, 0.25, 0.5, 0.75, 1.0])
    ax.set_yticklabels([0, 25, 50, 75, 100])
    ax.tick_params(direction='out', length=20, width=5)
    ax = change_box_to_arrow_axes(
        fig, ax,
        linewidth=5.0,
        overhang=0.0,
        xaxis_ypos=-0.05,
        yaxis_xpos=-0.1,
        x_offset=[-0.05, 0.1],
        y_offset=[-0.05, 0.2]
    )
    return fig, ax

def plot_force_length_curve():
    strain_range = np.linspace(0.5, 2, 1000)
    force_length_weight = force_length_curve_guassian(strain_range)
    passive_force_weight = passive_force_curve(strain_range)
    
    figsize = (19.2, 10.8)
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    ax.plot(strain_range, force_length_weight, color='C3', linewidth=8, label='active')
    ax.plot(strain_range, passive_force_weight, color='saddlebrown', linewidth=8, label='passive')
    ax.plot(strain_range, force_length_weight+passive_force_weight, color='black', linewidth=8, label='total')
    ax.legend(prop={"size":40}, frameon=False)
    ax.set_xlim(-0.1, 2.1)
    ax.tick_params(axis='x', labelsize=48)
    ax.tick_params(axis='y', labelsize=48)
    ax.set_yticks([0, 0.25, 0.5, 0.75, 1.0])
    ax.set_yticklabels([0, 25, 50, 75, 100])
    ax.tick_params(direction='out', length=20, width=5)
    ax = change_box_to_arrow_axes(
        fig, ax,
        linewidth=5.0,
        overhang=0.0,
        xaxis_ypos=-0.05,
        yaxis_xpos=-0.1,
        x_offset=[-0.05, 0.1],
        y_offset=[-0.05, 0.2]
    )
    return fig, ax

def plot_force_velocity_curve():
    strain_rate_range = np.linspace(-2, 2, 1000)
    force_velocity_weight = force_velocity_curve_tanh(strain_rate_range)
    figsize = (19.2, 10.8)
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    ax.plot(strain_rate_range, force_velocity_weight, color='C1', linewidth=8, label='active')
    ax.legend(prop={"size":40}, frameon=False)
    ax.set_xlim(-2.1, 2.1)
    ax.set_ylim(0, 2.1)
    ax.tick_params(axis='x', labelsize=48)
    ax.tick_params(axis='y', labelsize=48)
    ax.set_yticks([0, 0.5, 1.0, 1.5, 2.0])
    ax.set_yticklabels(['', 50, 100, 150, 200])
    ax.tick_params(direction='inout', length=20, width=5)
    ax = change_box_to_arrow_axes(
        fig, ax,
        linewidth=5.0,
        overhang=0.0,
        xaxis_ypos=0,
        yaxis_xpos=0,
        x_offset=[-0.05, 0.2],
        y_offset=[-0.05, 0.2]
    )
    ax.spines['left'].set_position(('data', 0))
    return fig, ax

def plot_hill_force_length_active_curve():
    
    f_a = 1
    f_v = 1.0
    stress_max = 280e3

    stretch_list = []
    stress_act_list = []
    for stretch in np.linspace(0.0, 2.0, 1001):
        
        f_l = force_length_curve_hill(np.array([stretch]))
        stress_act = f_a * stress_max * f_v * f_l
        stretch_list.append(stretch)
        stress_act_list.append(stress_act)
    
    stretch_array = np.array(stretch_list)
    stress_act_array = np.array(stress_act_list)
    plt.plot(stretch_array, stress_act_array/max(stress_act_array), "k.")

def generate_random_experiment_data(number_of_samples=100, deg=3):
    np.random.seed(314159265)
    sigma_stretch = 0.2
    mu_stretch = 1.0
    stretch_data = sigma_stretch * np.random.randn(number_of_samples) + mu_stretch
    # stretch_data = np.linspace(0, 2, number_of_samples)

    sigma_force_weight = 0.05
    force_weight_data = force_length_curve_hill(stretch_data)
    force_weight_data += sigma_force_weight * np.random.randn(number_of_samples)
    force_weight_data[force_weight_data < 0] = 0

    f_l_coefficients = np.polyfit(stretch_data, force_weight_data, deg=deg)
    result = "model function: "
    for i in range(deg+1):
        if i == deg:
            result += " %.2f " % f_l_coefficients[i]
        elif i == deg-1:
            result += " %.2f x +" % f_l_coefficients[i]
        else:
            result += " %.2f x^%d +" % (f_l_coefficients[i], deg-i)
    print(result)
    
    f_l = np.poly1d(f_l_coefficients) 
    stretch = np.linspace(0, 2.0, 401)
    # stretch = np.linspace(0.4, 1.9, 401)
    force_weight = f_l(stretch)
    # force_weight -= min(force_weight)
    # force_weight /= max(force_weight)
    force_weight[force_weight < 0] = 0

    figsize = (19.2, 10.8)
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)

    # ax.scatter(stretch_data, force_weight_data, s=100, color='C2', label='exp')
    ax.plot(stretch, force_weight, color='C1', linewidth=8, label='model')
    # ax.plot(stretch, force_length_curve_hill(stretch), color='C3', linewidth=8, label='Hill')
    ax.legend(prop={"size":40}, frameon=False)
    ax.set_xlim(-0.1, 2.1)
    ax.set_ylim(-0.1, 1.1)
    ax.tick_params(axis='x', labelsize=48)
    ax.tick_params(axis='y', labelsize=48)
    ax.set_yticks([0, 0.25, 0.5, 0.75, 1.0])
    ax.set_yticklabels([0, 25, 50, 75, 100])
    ax.tick_params(direction='out', length=20, width=5)
    ax = change_box_to_arrow_axes(
        fig, ax,
        linewidth=5.0,
        overhang=0.0,
        xaxis_ypos=-0.1,
        yaxis_xpos=-0.1,
        x_offset=[-0.05, 0.1],
        y_offset=[-0.05, 0.2]
    )
    return fig, ax

if __name__ == "__main__":
    import matplotlib.pyplot as plt
    # plot_force_length_active_curve()
    # plot_force_length_curve()
    # plot_force_velocity_curve()
    # plot_hill_force_length_active_curve()
    generate_random_experiment_data()
    plt.show()



from typing import Optional

from gymnasium import spaces, Env
from gymnasium.utils import seeding

from collections import defaultdict
import time
import copy

import numpy as np
from scipy.interpolate import interp1d

import elastica as el
from elastica.timestepper import extend_stepper_interface
from elastica._calculus import _isnan_check

from gym_softrobot import RENDERER_CONFIG
from gym_softrobot.config import RendererType
from gym_softrobot.envs.octopus.build import build_arm
from gym_softrobot.utils.custom_elastica.callback_func import (
    RodCallBack,
)
from gym_softrobot.utils.linalg import do_normalization
from gym_softrobot.utils.render.post_processing import plot_video
from gym_softrobot.utils.render.base_renderer import (
    BaseRenderer,
    BaseElasticaRendererSession,
)


class BaseSimulator(
    el.BaseSystemCollection, el.Constraints, el.Connections, el.Forcing, el.CallBacks
):
    pass


class ArmSingleEnv(Env):
    """
    Description:
    Source:
    Observation:
    Actions:
    Reward:
    Starting State:
    Episode Termination:
    Solved Requirements:
    """

    metadata = {"render.modes": ["rgb_array", "human"]}

    def __init__(
        self,
        final_time=10.0,
        time_step=7.0e-5,
        recording_fps=20,
        n_elems=50,
        n_action=7,
        control_penalty_coeff=0.001,
        config_generate_video=False,
        policy_mode="centralized",
    ):
        # Integrator type

        self.final_time = final_time
        self.time_step = time_step
        self.total_steps = int(self.final_time / self.time_step)
        self.recording_fps = recording_fps
        self.step_skip = int(1.0 / (recording_fps * time_step))
        self.control_penalty_coeff = control_penalty_coeff

        self.n_elems = n_elems
        self.n_seg = n_elems - 1
        self.policy_mode = policy_mode

        # Spaces
        self.n_action = n_action  # number of interpolation point (3 curvatures)
        action_size = (self.n_action,)
        action_low = np.ones(action_size) * (-22)
        action_high = np.ones(action_size) * (22)
        self.action_space = spaces.Box(
            action_low, action_high, shape=action_size, dtype=np.float32
        )
        self._observation_size = (
            25,
        )  # ((self.n_seg + (self.n_elems+1) * 4 + self.n_action + 2),) # 2 for target
        self.observation_space = spaces.Box(
            -np.inf, np.inf, shape=self._observation_size, dtype=np.float32
        )

        self.metadata = {}
        self.reward_range = 10.0
        self._prev_action = np.zeros(
            list(self.action_space.shape), dtype=self.action_space.dtype
        )

        # Configurations
        self.config_generate_video = config_generate_video

        # Rendering-related
        self.viewer = None
        self.renderer = None

        # Determinism
        self.seed()

        self.kappa_range = [-49.33508476187419, 49.33545827754751]
        self.sigma_range = [-0.041537734572300755, 0.1019431615063144]
        self.kappa_rate_range = [-21.063520620377012, 24.664591289161944]
        self.sigma_rate_range = [-0.08387293777925456, 0.06838264835333994]

    def seed(self, seed=None):
        # Deprecated in new gym
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def summary(
        self,
    ):
        print(
            f"""
        {self.final_time=}
        {self.time_step=}
        {self.total_steps=}
        {self.step_skip=}
        simulation time per action: {1.0/self.step_skip=}
        max number of action per episode: {self.total_steps / self.step_skip}

        {self.n_elems=}
        {self.action_space=}
        {self.observation_space=}
        {self.reward_range=}
        """
        )

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        return_info: bool = False,
        options: Optional[dict] = None,
    ):
        super().reset(seed=seed)
        self.simulator = BaseSimulator()

        self.shearable_rod = build_arm(
            self.simulator,
            self.n_elems,
        )

        # CallBack
        if self.config_generate_video:
            self.rod_parameters_dict = defaultdict(list)
            self.simulator.collect_diagnostics(self.shearable_rod).using(
                RodCallBack,
                step_skip=self.step_skip,
                callback_params=rod_parameters_dict,
            )

        """ Finalize the simulator and create time stepper """
        self.StatefulStepper = el.PositionVerlet()
        self.simulator.finalize()
        self.do_step, self.stages_and_updates = extend_stepper_interface(
            self.StatefulStepper, self.simulator
        )

        self.time = np.float64(0.0)
        self.counter = 0

        # Set Target
        self._target = np.array([1.0, 0.0])  # (2-0.5)*self.np_random.random(2) + 0.5
        # self._target /= np.linalg.norm(self._target) # I don't see why this is here
        # print(self._target)

        # Initial State
        rod = self.shearable_rod
        self.prev_kappa_state = copy.deepcopy(rod.kappa[0])
        self.prev_com_state = rod.compute_position_center_of_mass()[:2]
        state = self.get_state()

        # Preprocessing
        self.prev_dist_to_target = np.linalg.norm(
            self.shearable_rod.compute_position_center_of_mass()[:2] - self._target,
            ord=2,
        )
        # self.prev_cm_vel = self.shearable_rod.compute_velocity_center_of_mass()

        if return_info:
            return state, {}
        else:
            return state

    def get_state(self):
        # Build state
        rod = self.shearable_rod
        kappa_state = rod.kappa[0]
        kappa_rate_state = kappa_state - self.prev_kappa_state
        self.prev_kappa_state[...] = kappa_state

        mean_kappa_state = np.mean(kappa_state.reshape((7, 7)), axis=1)
        mean_kappa_rate_state = np.mean(kappa_rate_state.reshape((7, 7)), axis=1)

        com_state = rod.compute_position_center_of_mass()[:2]
        com_rate_state = com_state - self.prev_com_state
        self.prev_com_state[...] = com_state
        # pos_state1 = rod.position_collection[0] # x
        # pos_state2 = rod.position_collection[1] # y
        # vel_state1 = rod.velocity_collection[0] # x
        # vel_state2 = rod.velocity_collection[1] # y
        previous_action = self._prev_action.copy()
        target = self._target
        # state = np.hstack([
        #     mean_kappa_state, mean_kappa_rate_state,com_rate_state,# com_state, #pos_state1, pos_state2, vel_state1, vel_state2,
        #     previous_action, target]).astype(np.float32)
        normalized_kappa_state = self.normalize_state(
            mean_kappa_state, mean_kappa_rate_state
        )
        state = np.hstack(
            [
                normalized_kappa_state,
                com_rate_state,  # com_state, #pos_state1, pos_state2, vel_state1, vel_state2,
                previous_action,
                target,
            ]
        ).astype(np.float32)
        return state

    def normalize_state(self, kappa, kappa_rate):
        k = do_normalization(kappa, self.kappa_range)
        kr = do_normalization(kappa_rate, self.kappa_rate_range)
        return np.concatenate([k, kr])

    def set_action(self, action) -> None:
        self._prev_action[:] = action
        # action = np.concatenate([[0], action, [0]], axis=-1)
        action = interp1d(
            np.linspace(0, 1, self.n_action),  # added zero on the boundary
            action,
            kind="cubic",
            axis=-1,
        )(np.linspace(0, 1, self.n_seg))
        self.shearable_rod.rest_kappa[0, :] = action  # Planar curvature

    def step(self, action):
        rest_kappa = action  # alias

        """ Set intrinsic strains (set actions) """
        self.set_action(rest_kappa)

        """ Post-simulation """

        """ Run the simulation for one step """
        stime = time.perf_counter()
        for _ in range(self.step_skip):
            self.time = self.do_step(
                self.StatefulStepper,
                self.stages_and_updates,
                self.simulator,
                self.time,
                self.time_step,
            )
        etime = time.perf_counter()
        # print(f'{self.counter=}, {etime-stime}sec, {self.time=}')

        """ Done is a boolean to reset the environment before episode is completed """
        terminated = False
        truncated = False
        self.survive_reward = 0.0
        self.forward_reward = 0.0
        self.control_panelty = (
            self.control_penalty_coeff * np.square(rest_kappa.ravel()).mean()
        )
        # Position of the rod cannot be NaN, it is not valid, stop the simulation
        invalid_values_condition = (
            _isnan_check(
                np.concatenate(
                    [
                        self.shearable_rod.position_collection,
                        self.shearable_rod.velocity_collection,
                    ]
                )
            )
            or np.linalg.norm(self.shearable_rod.omega_collection) > 250
        )

        if invalid_values_condition:
            print(f" Nan detected in, exiting simulation now. {self.time=}")
            terminated = True
            truncated = True
            self.survive_reward = -1.0  # 50.0
        else:
            self.cm_pos = self.shearable_rod.compute_position_center_of_mass()[:2]
            dist_to_target = np.linalg.norm(self.cm_pos - self._target, ord=2)
            self.forward_velocity = self.prev_dist_to_target - dist_to_target
            self.forward_reward = (
                np.exp(-dist_to_target / 0.35) - 0.096
            )  # np.exp((self.cm_pos[0]-0.18)/0.35)-1#np.exp(self.forward_velocity)-1#(self.prev_dist_to_target - dist_to_target) * 10
            self.prev_dist_to_target = dist_to_target
            """ Goal """
            if dist_to_target < 0.1:
                self.survive_reward = 5.0
                terminated = True

        """ Time limit """
        timelimit = False
        if self.time > self.final_time:
            timelimit = True
            terminated = True

        reward = self.forward_reward - self.control_panelty + self.survive_reward
        # reward *= 10 # Reward scaling
        # print(f'{reward=:.3f}: {forward_reward=:.3f}, {control_panelty=:.3f}, {survive_reward=:.3f}')

        """ Return state:
            (1) current simulation time
            (2) current systems
            (3) a flag denotes whether the simulation runs correlectly
        """
        # systems = [self.shearable_rod]
        states = self.get_state()

        # Info
        info = {
            "time": self.time,
            "rod": self.shearable_rod,
            "TimeLimit.truncated": timelimit,
        }

        self.counter += 1

        return states, reward, terminated, truncated, info

    def save_data(self, filename_video, fps):
        if self.config_generate_video:
            filename_video = f"save/{filename_video}"
            plot_video(self.rod_parameters_dict, filename_video, margin=0.2, fps=fps)

    def render(self, mode="human", close=False):
        maxwidth = 800
        aspect_ratio = 3 / 4

        if self.viewer is None:
            from gym_softrobot.utils.render import pyglet_rendering

            self.viewer = pyglet_rendering.SimpleImageViewer(maxwidth=maxwidth)

        if self.renderer is None:
            # Switch renderer depending on configuration
            if RENDERER_CONFIG == RendererType.POVRAY:
                from gym_softrobot.utils.render.povray_renderer import Session
            elif RENDERER_CONFIG == RendererType.MATPLOTLIB:
                from gym_softrobot.utils.render.matplotlib_renderer import Session
            else:
                raise NotImplementedError("Rendering module is not imported properly")
            assert issubclass(Session, BaseRenderer), (
                "Rendering module is not properly subclassed"
            )
            assert issubclass(Session, BaseElasticaRendererSession), (
                "Rendering module is not properly subclassed"
            )
            self.viewer = pyglet_rendering.SimpleImageViewer(maxwidth=maxwidth)
            self.renderer = Session(width=maxwidth, height=int(maxwidth * aspect_ratio))
            self.renderer.add_rods(
                [self.shearable_rod]
            )  # TODO: maybe need add_rod instead
            self.renderer.add_point(self._target.tolist() + [0], 0.02)

        # POVRAY
        if RENDERER_CONFIG == RendererType.POVRAY:
            state_image = self.renderer.render(
                maxwidth, int(maxwidth * aspect_ratio * 0.7)
            )
            state_image_side = self.renderer.render(
                maxwidth // 2,
                int(maxwidth * aspect_ratio * 0.3),
                camera_param=("location", [0.0, 0.0, -0.5], "look_at", [0.0, 0, 0]),
            )
            state_image_top = self.renderer.render(
                maxwidth // 2,
                int(maxwidth * aspect_ratio * 0.3),
                camera_param=("location", [0.0, 0.3, 0.0], "look_at", [0.0, 0, 0]),
            )

            state_image = np.vstack(
                [state_image, np.hstack([state_image_side, state_image_top])]
            )
        elif RENDERER_CONFIG == RendererType.MATPLOTLIB:
            state_image = self.renderer.render()
        else:
            raise NotImplementedError("Rendering module is not imported properly")

        self.viewer.imshow(state_image)

        return state_image

    def close(self):
        if self.viewer:
            self.viewer.close()
            self.viewer = None
        if self.renderer:
            self.renderer.close()
            self.renderer = None



from typing import Optional, Any, Union, Dict

from gymnasium import spaces, Env
from gymnasium.utils import seeding

import time

import numpy as np
from scipy.interpolate import interp1d

from elastica import *
from elastica.timestepper import extend_stepper_interface
from elastica._calculus import _isnan_check

from gym_softrobot.envs.octopus.build_muscle_octopus import build_two_arms

from gym_softrobot.config import RendererType
from gym_softrobot import RENDERER_CONFIG
from gym_softrobot.utils.render.base_renderer import (
    BaseRenderer,
    BaseElasticaRendererSession,
)
from gym_softrobot.envs.octopus.controllable_constraint import ControllableFixConstraint


class BaseSimulator(BaseSystemCollection, Constraints, Connections, Forcing, CallBacks):
    pass


class ArmTwoEnv(Env):
    """
    Description:
        Decentralized
    Source:
    Observation:
    Actions:
    Reward:
    Starting State:
    Episode Termination:
    Solved Requirements:
    """

    metadata = {"render.modes": ["rgb_array"]}

    def __init__(
        self,
        final_time=5.0,
        time_step=5.0e-5,
        recording_fps=25,
        n_elems=20,
    ):
        # Integrator type
        self.final_time = final_time
        self.time_step = time_step
        self.total_steps = int(self.final_time / self.time_step)
        self.recording_fps = recording_fps
        self.step_skip = int(1.0 / (recording_fps * time_step))

        n_arm = 2
        self.n_arm = n_arm
        self.n_elems = n_elems
        self.n_seg = n_elems - 1
        self.n_sucker = 3
        self.sucker_location = [
            self.n_elems // (self.n_sucker * 2) * (2 * i + 1)
            for i in range(self.n_sucker)
        ]
        self.control_location = [0] + self.sucker_location + [self.n_elems - 1]

        # Spaces
        n_action = 9
        self.n_action = n_action
        # TODO: for non-HER, use decentral training shapes
        self.action_space = spaces.Box(
            0.0, 1.0, shape=(n_arm * n_action,), dtype=np.float32
        )

        shared_space = 3
        self.grid_size = 1
        self._observation_size = (
            n_arm * (self.n_seg * 2 + n_action + n_arm + shared_space),
        )
        self.observation_space = spaces.Box(
            -np.inf, np.inf, shape=self._observation_size, dtype=np.float32
        )

        self.metadata = {}
        self.reward_range = 100.0
        self._prev_action = np.zeros(
            list(self.action_space.shape), dtype=self.action_space.dtype
        )
        self._prev_kappa = np.zeros((n_arm, self.n_elems - 1), dtype=np.float32)

        # Configurations

        # Rendering-related
        self.viewer = None
        self.renderer = None

        # Determinism
        self.seed()

    def get_env_info(self):
        return dict(n_actions=self.n_action, n_agents=8)

    def seed(self, seed=None):
        # Deprecated in new gym
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        return_info: bool = False,
        options: Optional[dict] = None,
    ):
        super().reset(seed=seed)
        self.simulator = BaseSimulator()

        self.shearable_rods, self.rigid_rod, self.muscle_activations = build_two_arms(
            self.simulator,
            self.n_elems,
        )

        """ Controller Setup """
        constraint_ids = []
        for i in range(self.n_arm):
            each_arm_constraints = []
            for j in range(self.n_sucker):
                constraint_id = (
                    self.simulator.constrain(self.shearable_rods[i])
                    .using(
                        ControllableFixConstraint,
                        index=self.sucker_location[j],
                    )
                    .id()
                )
                each_arm_constraints.append(constraint_id)
            constraint_ids.append(each_arm_constraints)

        self.simulator.finalize()

        """ Sucker Controller Hook """
        self.sucker_controller = []
        for each_arm_constraints in constraint_ids:
            each_arm_sucker = []
            for constraint_id in each_arm_constraints:
                controllable_constraint = dict(self.simulator._constraints)[
                    constraint_id
                ]
                controller = controllable_constraint.get_controller
                controller.turn_on()
                each_arm_sucker.append(controller)
            self.sucker_controller.append(each_arm_sucker)

        """ Finalize the simulator and create time stepper """
        self.StatefulStepper = PositionVerlet()
        self.do_step, self.stages_and_updates = extend_stepper_interface(
            self.StatefulStepper, self.simulator
        )

        # """ Return
        #     (1) total time steps for the simulation step iterations
        #     (2) systems for controller design
        # """
        # systems = [self.shearable_rod]
        self.time = np.float64(0.0)
        self.counter = 0
        # self.bias=self.shearable_rod.compute_position_center_of_mass()[0].copy()

        # Set Target
        # self._target = self.np_random.uniform(-self.grid_size, self.grid_size, size=2).astype(np.float32)
        self._target = np.array([5, 0], dtype=np.float32)

        # Initial State
        state = self.get_state()

        if return_info:
            return state, {}
        else:
            return state

    def get_state(self):
        # Build state
        kappa_state = np.vstack([rod.kappa[0] for rod in self.shearable_rods])
        pos_state1 = np.vstack(
            [rod.position_collection[0] for rod in self.shearable_rods]
        )  # x
        pos_state2 = np.vstack(
            [rod.position_collection[1] for rod in self.shearable_rods]
        )  # y
        vel_state1 = np.vstack(
            [rod.velocity_collection[0] for rod in self.shearable_rods]
        )  # x
        vel_state2 = np.vstack(
            [rod.velocity_collection[1] for rod in self.shearable_rods]
        )  # y
        previous_action = self._prev_action.reshape([self.n_arm, self.n_action])

        shared_state = np.concatenate(
            [
                # self._target,  # 2
                # self.rigid_rod.position_collection[:, 0],  # 3
                self.rigid_rod.velocity_collection[:, 0],  # 3
                # self.rigid_rod.director_collection[:, :, 0].ravel(),  # 9: orientation
            ],
            dtype=np.float32,
        )
        observation_state = np.hstack(
            [
                kappa_state,
                self._prev_kappa,  # pos_state1, pos_state2, vel_state1, vel_state2,
                previous_action,
                np.eye(self.n_arm),
                np.repeat(shared_state[None, ...], self.n_arm, axis=0),
            ]
        ).astype(np.float32)
        self._prev_kappa[...] = kappa_state
        return np.nan_to_num(observation_state.ravel())

    def set_action(self, action) -> None:
        # Action: (8, n_action)
        scale = 1.0  # min(time / 0.02, 1)

        action = np.reshape(action, [self.n_arm, self.n_action])

        # Continuous action
        for i in range(self.n_arm):
            # f2=interp1d(self.control_location, action[i,:3], kind='cubic')
            sucker_activation = action[i, :3]
            LM_activation = action[i, 3:6]
            TM_activation = action[i, 6:]

            for j in range(self.n_sucker):
                self.sucker_controller[i][j].reduction_ratio = sucker_activation[j]

            LM_activation = LM_activation - 0.5
            LM1_activation = np.max((LM_activation, [0.0] * 3), axis=0)
            LM2_activation = abs(np.min((LM_activation, [0.0] * 3), axis=0))
            apply_LM1 = interp1d(
                self.control_location, [0] + list(LM1_activation) + [0], kind="cubic"
            )(range(self.n_elems - 1))
            apply_LM2 = interp1d(
                self.control_location, [0] + list(LM2_activation) + [0], kind="cubic"
            )(range(self.n_elems - 1))
            apply_TM = interp1d(
                self.control_location, [0] + list(TM_activation) + [0], kind="cubic"
            )(range(self.n_elems - 1))

            self.muscle_activations[i][0].set_activation(apply_LM1)
            self.muscle_activations[i][1].set_activation(apply_LM2)
            self.muscle_activations[i][2].set_activation(apply_TM)

        # update previous action
        self._prev_action = action

    def step(self, action):
        """Set intrinsic strains (set actions)"""
        self.set_action(action)

        """ Post-simulation """
        xposbefore = self.rigid_rod.position_collection[0:2, 0].copy()

        """ Run the simulation for one step """
        stime = time.perf_counter()
        for _ in range(self.step_skip):
            self.time = self.do_step(
                self.StatefulStepper,
                self.stages_and_updates,
                self.simulator,
                self.time,
                self.time_step,
            )
            # Debug
            """
            invalid_values_condition = _isnan_check(np.concatenate(
                [rod.position_collection for rod in self.shearable_rods] +
                [rod.velocity_collection for rod in self.shearable_rods]
                ))
            if invalid_values_condition == True:
                print(f" Nan detected in, exiting simulation now. {self.time=}")
                self.render()
                break
            else:
                self.render()
            """
        etime = time.perf_counter()
        states = self.get_state()

        """ Done is a boolean to reset the environment before episode is completed """
        terminated = False
        truncated = False
        survive_reward = 0.0
        forward_reward = 0.0
        control_cost = 0.0  # 0.5 * np.square(action.ravel()).mean()
        bending_energy = (
            0.0  # sum([rod.compute_bending_energy() for rod in self.shearable_rods])
        )
        shear_energy = (
            0.0  # sum([rod.compute_shear_energy() for rod in self.shearable_rods])
        )
        # Position of the rod cannot be NaN, it is not valid, stop the simulation
        invalid_values_condition = _isnan_check(
            np.concatenate(
                [rod.position_collection for rod in self.shearable_rods]
                + [rod.velocity_collection for rod in self.shearable_rods]
            )
        )

        if invalid_values_condition == True:
            # print(f" Nan detected in, exiting simulation now. {self.time=}")
            terminated = True
            truncated = True
            survive_reward = -5.0
        else:
            xposafter = self.rigid_rod.position_collection[0:2, 0]
            forward_reward = (
                np.linalg.norm(self._target - xposbefore)
                - np.linalg.norm(self._target - xposafter)
            ) * 1e2

            # forward_reward = self.compute_reward(
            #        self.rigid_rod.position_collection[:2,0],
            #        self._target,
            #        None)
            if np.linalg.norm(self._target - xposafter) < 0.2:
                survive_reward = 5
                terminated = True

        # print(self.rigid_rods.position_collection)
        # print(f'{self.counter=}, {etime-stime}sec, {self.time=}')
        if not terminated and self.time > self.final_time:
            forward_reward -= np.linalg.norm(self._target - xposafter)
            terminated = True

        reward = forward_reward - control_cost + survive_reward - bending_energy
        # reward *= 10 # Reward scaling
        # print(f'{reward=:.3f}, {forward_reward=:.3f}, {control_cost=:.3f}, {survive_reward=:.3f}, {bending_energy=:.3f}') #, {shear_energy=:.3f}')

        """ Return state:
            (1) current simulation time
            (2) current systems
            (3) a flag denotes whether the simulation runs correlectly
        """
        # systems = [self.shearable_rod]

        # Info
        info = {"time": self.time, "rods": self.shearable_rods, "body": self.rigid_rod}
        if np.isnan(reward):
            # TODO: Not sure why this happens
            reward = -5
            terminated = True
            truncated = True
        reward = min(self.reward_range, reward)

        self.counter += 1

        return states, reward, terminated, truncated, info

    def compute_reward(
        self,
        achieved_goal: Union[int, np.ndarray],
        desired_goal: Union[int, np.ndarray],
        _info: Optional[Dict[str, Any]] = None,
    ) -> np.float32:
        eps = 0.01
        dist = np.linalg.norm(achieved_goal - desired_goal, axis=-1)
        return -(dist > eps).astype(np.float32)

    def render(self, mode="human", close=False) -> Optional[np.ndarray]:
        maxwidth = 800
        aspect_ratio = 3 / 4

        if self.viewer is None:
            from gym_softrobot.utils.render import pyglet_rendering

            self.viewer = pyglet_rendering.SimpleImageViewer(maxwidth=maxwidth)

        if self.renderer is None:
            # Switch renderer depending on configuration
            if RENDERER_CONFIG == RendererType.POVRAY:
                from gym_softrobot.utils.render.povray_renderer import Session
            elif RENDERER_CONFIG == RendererType.MATPLOTLIB:
                from gym_softrobot.utils.render.matplotlib_renderer import Session
            else:
                raise NotImplementedError("Rendering module is not imported properly")
            assert issubclass(Session, BaseRenderer), (
                "Rendering module is not properly subclassed"
            )
            assert issubclass(Session, BaseElasticaRendererSession), (
                "Rendering module is not properly subclassed"
            )
            self.renderer = Session(width=maxwidth, height=int(maxwidth * aspect_ratio))
            self.renderer.add_rods(self.shearable_rods)
            self.renderer.add_rigid_body(self.rigid_rod)
            # self.renderer.add_point(self._target.tolist()+[0], 0.20)

        # POVRAY
        if RENDERER_CONFIG == RendererType.POVRAY:
            state_image = self.renderer.render(
                maxwidth, int(maxwidth * aspect_ratio * 0.7)
            )
            state_image_side = self.renderer.render(
                maxwidth // 2,
                int(maxwidth * aspect_ratio * 0.3),
                camera_param=("location", [0.0, 0.0, -0.5], "look_at", [0.0, 0, 0]),
            )
            state_image_top = self.renderer.render(
                maxwidth // 2,
                int(maxwidth * aspect_ratio * 0.3),
                camera_param=("location", [0.0, 0.3, 0.0], "look_at", [0.0, 0, 0]),
            )

            state_image = np.vstack(
                [state_image, np.hstack([state_image_side, state_image_top])]
            )
        elif RENDERER_CONFIG == RendererType.MATPLOTLIB:
            state_image = self.renderer.render()
        else:
            raise NotImplementedError("Rendering module is not imported properly")

        self.viewer.imshow(state_image)

        return state_image

    def close(self):
        if self.viewer:
            self.viewer.close()
            self.viewer = None
        if self.renderer:
            self.renderer.close()
            self.renderer = None



"""
Created on Oct.19, 2020
@author: Heng-Sheng (Hanson) Chang
"""

from gym_softrobot.utils.actuation.objects.object import *
from gym_softrobot.utils.actuation.objects.point import *
from gym_softrobot.utils.actuation.objects.cylinder import *



__doc__ = """
Module contains elastica interface to create soft pendulum.
"""

from typing import Optional

import numpy as np

from elastica import *
from elastica.boundary_conditions import ConstraintBase


_PENDULUM_PROPERTIES = {  # default parameters
    # Arm properties
    "youngs_modulus": 1e6,
    "density": 1000.0,
    "nu": 50,
}
_DEFAULT_SCALE_LENGTH = {
    "base_length": 1.0,
    "base_radius": 0.05,
}


def build_soft_pendulum(
    simulator,
    n_elem,
    point_force,
    np_random,
    time_step,
    override_params: Optional[dict] = None,
):
    """Import default parameters (overridable)"""
    param = _PENDULUM_PROPERTIES.copy()  # Always copy parameter for safety
    if isinstance(override_params, dict):
        param.update(override_params)
    """ Import default parameters (non-overridable) """
    pendulum_scale_param = _DEFAULT_SCALE_LENGTH.copy()

    """ Set pendulum """

    start = np.zeros((3,))
    theta = np.deg2rad(
        90 + (np_random.random() - 0.5) * 10
    )  # Change this if you want to start with different initial condition.
    direction = np.array([1.0 * np.cos(theta), 1.0 * np.sin(theta), 0.0])
    normal = np.array([1.0 * np.sin(theta), -1.0 * np.cos(theta), 0.0])
    binormal = np.cross(direction, normal)

    shearable_rod = CosseratRod.straight_rod(
        n_elem,
        start,
        direction,
        normal,
        **pendulum_scale_param,
        **param,
    )
    simulator.append(shearable_rod)

    # Add boundary conditions
    class PendulumBoundaryConditions(ConstraintBase):
        def __init__(self, fixed_position, fixed_directors, **kwargs):
            super().__init__(**kwargs)
            self.fixed_position = fixed_position
            self.fixed_directors = fixed_directors

        def constrain_values(self, rod, time):
            rod.position_collection[1:, 0] = self.fixed_position[1:]
            rod.director_collection[0, :, 0] = self.fixed_directors[0, :]
            rod.director_collection[2, :, 0] = self.fixed_directors[2, :]

        def constrain_rates(self, rod, time):
            rod.velocity_collection[1:, 0] = 0
            rod.omega_collection[0, 0] = 0
            rod.omega_collection[2, 0] = 0

    simulator.constrain(shearable_rod).using(
        PendulumBoundaryConditions,
        constrained_position_idx=(0,),
        constrained_director_idx=(0,),
    )

    # Add gravity
    gravitational_acc = -9.80665
    simulator.add_forcing_to(shearable_rod).using(
        GravityForces, acc_gravity=np.array([0.0, gravitational_acc, 0.0])
    )

    # Apply external forces to the base of the rod.
    class PendulumPointForces(NoForces):
        def __init__(self, point_force):
            super(PendulumPointForces, self).__init__()

            self.point_force = point_force

        def apply_forces(self, system, time: np.float = 0.0):
            system.external_forces[0, 0] = self.point_force

    simulator.add_forcing_to(shearable_rod).using(
        PendulumPointForces, point_force=point_force
    )

    # add damping
    damping_constant = 2e-3
    simulator.dampen(shearable_rod).using(
        AnalyticalLinearDamper,
        damping_constant=damping_constant,
        time_step=time_step,
    )

    return shearable_rod



"""
Created on Fri Apr. 17, 2020
@author: Heng-Sheng (Hanson) Chang
"""

import numpy as np
from numba import njit

from gym_softrobot.utils.actuation.objects.object import Constraint, Target

from elastica._rotations import _inv_rotate

class CylinderConstraint(Constraint):
    def __init__(self, center, normal, radius, height, constraint_cost_weight):
        Constraint.__init__(self, 'cylinder', constraint_cost_weight)
        self.center = np.array(center)
        self.normal = np.array(normal)
        self.normal = self.normal/np.linalg.norm(self.normal)
        self.radius = radius
        self.height = height
        self.indicator = None
        self.closest_position = None

    def running_cost(self, position, radii):
        self.indicator = np.zeros(position.shape[1])
        self.closest_position = np.zeros(position.shape)
        indicator(self.center, self.normal, self.radius, position, radii, self.indicator, self.closest_position)
        constraint_running_cost(self.closest_position, position, self.indicator, self.cost, self.constraint_cost_weight)
        return

    def running_cost_partial_position(self, position):
        return constraint_running_cost_partial_position(self.closest_position, position, self.indicator, self.constraint_cost_weight)

    def calculate_cost(self, position, radii):
        self.cost[0] = 0
        self.running_cost(position, radii)
        return

@njit(cache=True)
def indicator(center, normal, radius, position, radii, indicator_array, closest_position):
    for i in range(indicator_array.shape[0]):
        delta_position = position[:, i] - center
        normal_delta_position = np.dot(normal, delta_position) * normal
        parallel_delta_position = delta_position - normal_delta_position
        temp_center = center + normal_delta_position

        vector_length = np.linalg.norm(parallel_delta_position)
        equivalent_distance = radii[i] + radius
        value = vector_length - equivalent_distance
        closest_position[:, i] = temp_center + (equivalent_distance/vector_length)*parallel_delta_position
        # closest_position[:, i] += center
        if value > 0: # outside
            indicator_array[i] = 1
        elif value < 0: # inside
            indicator_array[i] = -1
        else:
            indicator_array[i] = 0
    return

@njit(cache=True)
def constraint_running_cost(closest_position, position, indicator_array, cost, constraint_cost_weight):
    for i in range(indicator_array.shape[0]):
        if indicator_array[i] < 0:
            cost[0] += 0.5*constraint_cost_weight * \
                ((position[0, i]-closest_position[0, i])**2 + \
                 (position[1, i]-closest_position[1, i])**2 + \
                 (position[2, i]-closest_position[2, i])**2)
    return

@njit(cache=True)
def constraint_running_cost_partial_position(closest_position, position, indicator_array, constraint_cost_weight):
    internal_force = np.zeros(position.shape)
    for i in range(indicator_array.shape[0]):
        if indicator_array[i] < 0:
            internal_force[:, i] = constraint_cost_weight * (closest_position[:, i]-position[:, i])
    return internal_force


class CylinderTarget(Target, CylinderConstraint):
    def __init__(self, position, normal, radius, heights, constraint_cost_weight, target_running_cost_weight, target_cost_weight=0):
        Target.__init__(self, 'cylinder', position, None, target_cost_weight)
        assert (heights[0]*heights[1]) <= 0, \
            "one of the two entries in heights should be nonnegative and the other should be nonpositive."
        assert not ((heights[0] == 0) and (heights[1] == 0)), \
            "two entries in heights can not all be zero."
        self.heights = dict(
            upper=max(heights[0], heights[1]),
            lower=-min(heights[0], heights[1])
            )
        self.normal = np.array(normal)
        self.normal = self.normal/np.linalg.norm(self.normal)
        self.target_running_cost_weight = np.array(target_running_cost_weight)
        CylinderConstraint.__init__(self,
            self.position - self.heights['lower']*self.normal,
            self.normal, radius, self.heights['upper']+self.heights['lower'], constraint_cost_weight)

    def terminal_cost(self, end_position):
        self.cost[0] += 0.5 * self.target_cost_weight * np.linalg.norm(self.position - end_position)**2
        return

    def terminal_cost_partial_end_state(self, end_position, end_director=None):
        terminal_cost_partial_end_position = self.target_cost_weight * (end_position - self.position)

        if not (self.director == None):
            director_collection = np.zeros((3, 3, 2))
            director_collection[:, :, 0] = end_director
            director_collection[:, :, 1] = self.director
            rotate_axis_and_angle = _inv_rotate(director_collection)
            terminal_cost_partial_end_director = self.target_cost_weight * rotate_axis_and_angle[:, 0]
            return terminal_cost_partial_end_position, terminal_cost_partial_end_director

        return terminal_cost_partial_end_position, None
        
    def terminal_cost_partial_end_position(self, end_position):
        return self.target_cost_weight * (end_position - self.position)

    def running_cost(self, position, radii):
        self.indicator = np.zeros(position.shape[1])
        self.closest_position = np.zeros(position.shape)
        indicator(self.center, self.normal, self.radius, position, radii, self.indicator, self.closest_position)
        target_running_cost(self.closest_position, position, self.indicator, self.cost, self.constraint_cost_weight, self.target_running_cost_weight)
        return

    def running_cost_partial_position(self, position):
        return target_running_cost_partial_position(self.closest_position, position, self.indicator, self.constraint_cost_weight, self.target_running_cost_weight)

    def calculate_cost(self, end_position, position, radii):
        self.cost[0] = 0
        self.running_cost(position, radii)
        if not (self.target_cost_weight == 0):
            self.terminal_cost(end_position)
        return

@njit(cache=True)
def target_running_cost(closest_position, position, indicator_array, cost, constraint_cost_weight, target_running_cost_weight):
    for i in range(indicator_array.shape[0]):
        if indicator_array[i] < 0:
            cost[0] += 0.5*constraint_cost_weight * \
                ((position[0, i]-closest_position[0, i])**2 + \
                 (position[1, i]-closest_position[1, i])**2 + \
                 (position[2, i]-closest_position[2, i])**2)
        if indicator_array[i] > 0:
            cost[0] += 0.5*target_running_cost_weight[i] * \
                ((position[0, i]-closest_position[0, i])**2 + \
                 (position[1, i]-closest_position[1, i])**2 + \
                 (position[2, i]-closest_position[2, i])**2)
    return

@njit(cache=True)
def target_running_cost_partial_position(closest_position, position, indicator_array, constraint_cost_weight, target_running_cost_weight):
    internal_force = np.zeros(position.shape)
    for i in range(indicator_array.shape[0]):
        if indicator_array[i] < 0:
            internal_force[:, i] = constraint_cost_weight * (closest_position[:, i]-position[:, i])
        if indicator_array[i] > 0:
            internal_force[:, i] = target_running_cost_weight[i] * (closest_position[:, i]-position[:, i])
    return internal_force



"""
Created on Sep. 23, 2021
@author: Heng-Sheng (Hanson) Chang
"""
import numpy as np
from numba import njit
import matplotlib.colors as mcolors
from numpy.core.shape_base import block

default_colors = mcolors.TABLEAU_COLORS
# default_colors['tab:blue']    # muted blue
# default_colors['tab:orange']  # safety orange
# default_colors['tab:green']   # cooked asparagus green
# default_colors['tab:red']     # brick red
# default_colors['tab:purple']  # muted purple
# default_colors['tab:brown']   # chestnut brown
# default_colors['tab:pink']    # raspberry yogurt pink
# default_colors['tab:gray']    # middle gray
# default_colors['tab:olive']   # curry yellow-green
# default_colors['tab:cyan']    # blue-teal

base_colors = mcolors.BASE_COLORS
# base_colors['b']              # blue
# base_colors['g']              # green
# base_colors['r']              # red
# base_colors['c']              # cyan
# base_colors['m']              # magenta
# base_colors['y']              # yellow
# base_colors['k']              # black
# base_colors['w']              # white
    
default_label_fontsize = 15
paper_label_fontsize = 48
paper_linewidth = 5

rod_color = base_colors['m']
algo_color = base_colors['g']

@njit(cache=True)
def process_position(position, offset, rotation):
    blocksize = position.shape[1]
    output_position = np.zeros((3, blocksize))
    for n in range(blocksize):
        for i in range(3):
            for j in range(3):
                output_position[i, n] += (
                    rotation[i, j] * (position[j, n] - offset[j])
                )
    return output_position

@njit(cache=True)
def process_director(director, rotation):
    blocksize = director.shape[2]
    output_director = np.zeros((3, 3, blocksize))
    for n in range(blocksize):
        for i in range(3):
            for j in range(3):
                for k in range(3):
                    output_director[i, j, n] += (
                        rotation[i, k] * director[k, j, n]
                    )
    return output_director

def change_box_to_arrow_axes(
    fig, ax, linewidth=1.0, overhang=0.0,
    xaxis_ypos=0, yaxis_xpos=0,
    x_offset=[0, 0], y_offset=[0, 0]
):
    for spine in ax.spines.values():
        spine.set_visible(False)

    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()

    # get width and height of axes object to compute
    # matching arrowhead length and width
    dps = fig.dpi_scale_trans.inverted()
    bbox = ax.get_window_extent().transformed(dps)
    width, height = bbox.width, bbox.height

    # manual arrowhead width and length
    hw = 1./40.*(ymax-ymin)
    hl = 1./40.*(xmax-xmin)
    lw = linewidth # axis line width
    ohg = overhang # arrow overhang

    # compute matching arrowhead length and width
    yhw = hw/(ymax-ymin)*(xmax-xmin)* height/width
    yhl = hl/(xmax-xmin)*(ymax-ymin)* width/height

    # draw x and y axis
    ax.arrow(xmin+x_offset[0], xaxis_ypos, xmax-xmin+x_offset[1], 0, fc='k', ec='k', lw = lw, 
                head_width=hw, head_length=hl, overhang = ohg, 
                length_includes_head= True, clip_on = False) 

    ax.arrow(yaxis_xpos, ymin+y_offset[0], 0, ymax-ymin+y_offset[1], fc='k', ec='k', lw = lw, 
                head_width=yhw, head_length=yhl, overhang = ohg, 
                length_includes_head= True, clip_on = False)
    return ax



__all__ = ["ConvertToPyMarlEnv"]

import typing
from typing import List, Iterable, Union, Optional
import numpy as np

import gym


class ConvertToPyMarlEnv:
    """ Convert environment to PyMarl multi-agent environments.
    Template from: oxwhirl/pymar/src/envs/multiagentenv.py

    Available benchmark algorithms: 

    - `oxwhirl/PyMarl <https://github.com/oxwhirl/pymarl>`_
    - `uoe-agents/EPyMarl <https://github.com/uoe-agents/epymarl>`_

    The wrapper to convert gym_softrobot environment (that is MA-compatible)
    to 'multiagentenv' that is compatible to PyMARL.
    The purpose is to run benchmark study with standard CTDE algorithms, such
    as QMIX, COMA, etc. Here is the example snippet::

        env = ConvertToPyMarlEnv(ma_env=gym.make("OctoCrawl-v0"))
    """

    def __init__(self, ma_env: gym.Env, *args, **kwargs):
        """

        Parameters
        ----------
        ma_env : gym.Env
            Multi-agent compatible environment
            In gym-softrobot, environment with 'multiagent' tag in meta data
            must be true.
        """
        super().__init__(*args, **kwargs)

        # Register environment
        assert (
            "PyMARL" in ma_env.metadata["multiagent"]
        ), f"The given environment is not multi-agent compatible"
        self.env = ma_env

    @property
    def n_agents(self):
        return self.env.n_agent

    def step(self, actions):
        """Returns reward, terminated, info"""
        self.observation, reward, done, info = self.env.step(actions.ravel())
        return reward, done, info

    def get_obs(self):
        """Returns all agent observations in a list"""
        return [self.get_obs_agent(i) for i in range(self.n_agents)]

    def get_obs_agent(self, agent_id):
        """Returns observation for agent_id."""
        obs_size = self.get_obs_size()
        return self.observation[agent_id * obs_size : (agent_id + 1) * obs_size]

    def get_obs_size(self):
        """Returns the shape of the observation

        The observation in regular gym_softrobot is given as (n * state_space)
        """
        obs_size = self.env.observation_space.shape[0] // self.n_agents
        return obs_size

    def get_state(self):
        """return global state.

        Notes
        -----
        Ideally, this function should not be used during decentralized execution.
        """
        return self.env.get_shared_state()

    def get_state_size(self):
        """Returns the shape of the state"""
        return self.env.shared_space

    def get_avail_actions(self) -> List:
        """Returns the available actions of all agents in a list."""
        return [
            self.get_aviala_agent_actions(agent_id) for agent_id in range(self.n_agents)
        ]

    def get_avail_agent_actions(self, agent_id):
        """Returns the available actions for agent_id"""
        raise NotImplementedError

    def get_total_actions(self):
        """Returns the total number of actions an agent could ever take"""
        # TODO: This is only suitable for a discrete 1 dimensional action space for each agent
        raise NotImplementedError

    def get_env_info(self):
        env_info = {
            "state_shape": self.get_state_size(),
            "obs_shape": self.get_obs_size(),
            "n_actions": self.get_total_actions(),
            "n_agents": self.n_agents,
            "episode_limit": self.episode_limit,
        }
        return env_info

    def save_replay(self):
        # TODO: Low priority
        raise NotImplementedError

    # Implemented in original environment
    def close(self):
        self.env.close()

    def seed(self, seed=None):
        self.env.seed(seed=seed)

    def render(self, **kwargs):
        self.env.render(**kwargs)

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        return_info: bool = False,
        options: Optional[dict] = None,
    ):
        """Returns initial observations and states"""
        self.observation = self.env.reset(**kwargs)
        return self.observation, self.get_state()



"""
Created on Jan. 07, 2021
@author: Heng-Sheng (Hanson) Chang
"""

from gym_softrobot.utils.actuation.actuations.muscles.muscle import *
from gym_softrobot.utils.actuation.actuations.muscles.oblique_muscle import *
from gym_softrobot.utils.actuation.actuations.muscles.longitudinal_muscle import *
from gym_softrobot.utils.actuation.actuations.muscles.transverse_muscle import *



import numpy as np
from matplotlib import pyplot as plt
from matplotlib.colors import to_rgb
from tqdm import tqdm
import matplotlib.animation as manimation


def plot_video(
    plot_params_list: list, video_name="video.mp4", margin=0.2, fps=15, plot3d=False
):

    t = np.array(plot_params_list[0]["time"])
    pos_list=[]
    for plot_params in plot_params_list:
        positions_over_time = np.array(plot_params["position"])
        pos_list.append(positions_over_time)
    total_time = int(np.around(t[..., -1], 1))
    total_frames = fps * total_time
    dump = round(len(t) / total_frames)
    print("creating video -- this can take a few minutes")
    FFMpegWriter = manimation.writers["ffmpeg"]
    metadata = dict(title="Movie Test", artist="Matplotlib", comment="Movie support!")
    writer = FFMpegWriter(fps=fps, metadata=metadata)
    fig = plt.figure()
    with writer.saving(fig, video_name, dpi=300):
        for time in range(1, len(t), dump):
            fig.clf()
            if plot3d:
                ax = plt.axes(projection="3d")
            else:
                ax = plt.axes()
            for positions_over_time in pos_list:
                x = positions_over_time[time][0]
                y = positions_over_time[time][1]
                z = positions_over_time[time][2]

                if plot3d:
                    ax.plot3D(x, y, z)
                    ax.set_xlim(0, 4)
                    ax.set_ylim(-1, 1)
                    ax.set_zlim(-0.05, 0.05)
                    ax.set_xlabel("x")
                    ax.set_ylabel("y")
                    ax.set_zlabel("z")
                else:
                    ax.plot(x, y)
                    ax.set_xlim(-0.5, 2.5)
                    ax.set_ylim(-0.5, 2.5)
                    ax.set_xlabel("x")
                    ax.set_ylabel("y")
                    ax.set_aspect("equal")
            writer.grab_frame()
    plt.close(fig)


def compute_projected_velocity(plot_params: dict, period):

    time_per_period = np.array(plot_params["time"]) / period
    avg_velocity = np.array(plot_params["avg_velocity"])
    center_of_mass = np.array(plot_params["center_of_mass"])

    # Compute rod velocity in rod direction. We need to compute that because,
    # after snake starts to move it chooses an arbitrary direction, which does not
    # have to be initial tangent direction of the rod. Thus we need to project the
    # snake velocity with respect to its new tangent and roll direction, after that
    # we will get the correct forward and lateral speed. After this projection
    # lateral velocity of the snake has to be oscillating between + and - values with
    # zero mean.

    # Number of steps in one period.
    period_step = int(period / (time_per_period[-1] - time_per_period[-2])) + 1
    number_of_period = int(time_per_period.shape[0] / period_step)
    # Center of mass position averaged in one period
    center_of_mass_averaged_over_one_period = np.zeros((number_of_period - 2, 3))
    for i in range(1, number_of_period - 1):
        # position of center of mass averaged over one period
        center_of_mass_averaged_over_one_period[i - 1] = np.mean(
            center_of_mass[(i + 1) * period_step : (i + 2) * period_step]
            - center_of_mass[(i + 0) * period_step : (i + 1) * period_step],
            axis=0,
        )

    # Average the rod directions over multiple periods and get the direction of the rod.
    direction_of_rod = np.mean(center_of_mass_averaged_over_one_period, axis=0)
    direction_of_rod /= np.linalg.norm(direction_of_rod, ord=2)

    # Compute the projected rod velocity in the direction of the rod
    velocity_mag_in_direction_of_rod = np.einsum(
        "ji,i->j", avg_velocity, direction_of_rod
    )
    velocity_in_direction_of_rod = np.einsum(
        "j,i->ji", velocity_mag_in_direction_of_rod, direction_of_rod
    )

    # Get the lateral or roll velocity of the rod after subtracting its projected
    # velocity in the direction of rod
    velocity_in_rod_roll_dir = avg_velocity - velocity_in_direction_of_rod

    # Compute the average velocity over the simulation, this can be used for optimizing snake
    # for fastest forward velocity. We start after first period, because of the ramping up happens
    # in first period.
    average_velocity_over_simulation = np.mean(
        velocity_in_direction_of_rod[period_step * 2 :], axis=0
    )

    return (
        velocity_in_direction_of_rod,
        velocity_in_rod_roll_dir,
        average_velocity_over_simulation[2],
        average_velocity_over_simulation[0],
    )




"""
Created on Oct. 19, 2020
@author: Heng-Sheng (Hanson) Chang
"""

import numpy as np
from numba import njit, jit

from elastica._linalg import _batch_matvec, _batch_cross, _batch_norm, _batch_matrix_transpose
from elastica._calculus import quadrature_kernel, difference_kernel

from elastica.external_forces import NoForces

@njit(cache=True)
def inverse_rigidity_matrix(matrix):
    inverse_matrix = np.zeros(matrix.shape)
    for i in range(inverse_matrix.shape[2]):
        inverse_matrix[:, :, i] = np.linalg.inv(matrix[:, :, i])
    return inverse_matrix
    
def strain_to_shear_and_curvature(strain):
    n_elems = int((strain.size + 3) / 6)
    shear = np.zeros((3, n_elems))
    curvature = np.zeros((3, n_elems-1))
    index = 0
    for i in range(3):
        shear[i, :] = strain[index:index+n_elems]
        index += n_elems
    for i in range(3):
        curvature[i] = strain[index:index+n_elems-1]
        index += (n_elems-1)
    return shear, curvature

def shear_and_curvature_to_strain(shear, curvature):
    strain = np.zeros(shear.size + curvature.size)
    n_elems = shear.shape[1]
    index = 0
    for i in range(3):
        strain[index:index+n_elems] = shear[i, :]
        index += n_elems
    for i in range(3):
        strain[index:index+n_elems-1] = curvature[i]
        index += (n_elems-1)
    return strain

@njit(cache=True)
def _lab_to_material(directors, lab_vectors):
    return _batch_matvec(directors, lab_vectors)

@njit(cache=True)
def _material_to_lab(directors, material_vectors):

    blocksize = material_vectors.shape[1]
    output_vector = np.zeros((3, blocksize))

    for i in range(3):
        for j in range(3):
            for k in range(blocksize):
                output_vector[i, k] += (
                    directors[j, i, k] * material_vectors[j, k]
                )

    return output_vector

@njit(cache=True)
def average(vector_collection):

    blocksize = vector_collection.shape[1]-1
    output_vector = np.zeros((3, blocksize))

    for k in range(blocksize):
        for i in range(3):
            output_vector[i, k] = (vector_collection[i, k]+vector_collection[i, k+1])/2

    return output_vector

@njit(cache=True)
def average1D(vector_collection):

    blocksize = vector_collection.shape[0]-1
    output_vector = np.zeros(blocksize)

    for k in range(blocksize):
        output_vector[k] = (vector_collection[k]+vector_collection[k+1])/2
    return output_vector


@njit(cache=True)
def distance(position_collection, target_position):
    blocksize = position_collection.shape[1]
    distance_collection = np.zeros(blocksize)
    for k in range(blocksize):
        distance_collection[k] = (
            (position_collection[0, k]-target_position[0])**2 +
            (position_collection[1, k]-target_position[1])**2 +
            (position_collection[2, k]-target_position[2])**2
        )**0.5
    return distance_collection

@njit(cache=True)
def forward_path(dl, sigma, kappa, position_collection, director_collection, r0=np.zeros(3), Q0=np.array([[0,0,-1],[0,1,0],[1,0,0]])):
    _, voronoi_dilatation = calculate_dilatation(sigma)
    shear = sigma_to_shear(sigma)
    curvature = kappa_to_curvature(kappa, voronoi_dilatation)

    # position_collection[:,0] = r0
    # director_collection[:,:,0] = Q0
    for i in range(kappa.shape[1]):
        # next_position(
        #     director_collection[:, :, i],
        #     (sigma[:, i]+np.array([0, 0, 1]))*dl,
        #     position_collection[:, i:i+2]
        #     )
        next_position(
            director_collection[:, :, i],
            shear[:, i] * dl,
            position_collection[:, i:i+2]
            )
        # next_director(
        #     voronoi_dilatation[i]*curvature[:, i]*dl,
        #     director_collection[:, :, i:i+2]
        #     )
        # next_director(
        #     kappa[:, i] * dl,
        #     director_collection[:, :, i:i+2]
        #     )
        next_director(
            kappa[:, i] * dl,
            director_collection[:, :, i:i+2]
            )

    next_position(
        director_collection[:, :, -1],
        shear[:, -1] * dl,
        position_collection[:, -2:]
        )
    pass

@njit(cache=True)
def next_position(director, delta, positions):
    positions[:, 1] = positions[:, 0]
    for index_i in range(3):
        for index_j in range(3):
            positions[index_i, 1] += director[index_j, index_i] * delta[index_j]
    return

@njit(cache=True)
def next_director(rotation, directors):
    Rotation = get_rotation_matrix(rotation)
    for index_i in range(3):
        for index_j in range(3):
            directors[index_i, index_j, 1] = 0
            for index_k in range(3):
                directors[index_i, index_j, 1] += (
                    Rotation[index_k, index_i] * directors[index_k, index_j, 0]
                )
    return

@njit(cache=True)
def get_rotation_matrix(rotation):
    Rotation = np.identity(3)
    angle = np.sqrt(rotation[0]**2+rotation[1]**2+rotation[2]**2)
    if angle == 0:
        return Rotation
    axis = rotation/angle
    K = np.zeros((3, 3))
    K[2, 1] = axis[0]
    K[1, 2] = - axis[0]
    K[0, 2] = axis[1]
    K[2, 0] = - axis[1]
    K[1, 0] = axis[2]
    K[0, 1] = - axis[2]

    K2 = np.zeros((3, 3))
    K2[0, 0] = -(axis[1]*axis[1]+axis[2]*axis[2])
    K2[1, 1] = -(axis[2]*axis[2]+axis[0]*axis[0])
    K2[2, 2] = -(axis[0]*axis[0]+axis[1]*axis[1])
    K2[0, 1] = axis[0]*axis[1]
    K2[1, 0] = axis[0]*axis[1]
    K2[0, 2] = axis[0]*axis[2]
    K2[2, 0] = axis[0]*axis[2]
    K2[1, 2] = axis[1]*axis[2]
    K2[2, 1] = axis[1]*axis[2]

    Rotation[:, :] += np.sin(angle)*K + (1-np.cos(angle))*K2
    return Rotation

@njit(cache=True)
def backward_path(dl, director, shear, ns, ms, n1, m1, internal_force, internal_couple):
    n = np.zeros(internal_force.shape)
    n[:, -1] = n1
    for i in range(n.shape[1]-1):
        n[:, -1-i-1] = n[:, -1-i] - (ns[:, -1-i]+ns[:, -1-i-1])*dl/2
    internal_force[:, :] = _batch_matvec(director, n)

    m = np.zeros(internal_force.shape)
    m[:, -1] = m1
    ms[:, :] = -_batch_cross(_batch_matvec(_batch_matrix_transpose(director), shear), n)
    for i in range(m.shape[1]-1):
        m[:, -1-i-1] = m[:, -1-i] - (ms[:, -1-i]+ms[:, -1-i-1])*dl/2
    internal_couple[:, :] = quadrature_kernel(_batch_matvec(director, m))[:, 1:-1]
    return

# @njit(cache=True)
# def update_deformation(internal_force, internal_couple, shear_matrix, bend_matrix, 
#             step_strain, step_curvature, strain, curvature, next_strain, next_curvature):
#     dilatation, voronoi_dilatation = calculate_dilatation(strain)
#     intrinsic_strain = np.array([0, 0, 1])
#     for i in range(next_strain.shape[1]):
#         for index_i in range(3):
#             next_strain[index_i, i] = strain[index_i, i] \
#                 + step_strain[index_i, i] * (internal_force[index_i, i] - (strain[index_i, i] - intrinsic_strain[index_i])*(shear_matrix[index_i, index_i, i]/dilatation[i]))
#         if next_strain[2, i] < 0:
#             next_strain[2, i] = 0

#     for i in range(next_curvature.shape[1]):
#         for index_i in range(3):
#             next_curvature[index_i, i] = curvature[index_i, i] \
#                 + step_curvature[index_i, i] * (internal_couple[index_i, i] - curvature[index_i, i]*(bend_matrix[index_i, index_i, i]/(voronoi_dilatation[i]**2)))
#     return

@njit(cache=True)
def calculate_dilatation(sigma):
    shear = sigma.copy()
    for i in range(shear.shape[1]):
        shear[2, i] += 1
    dilatation = _batch_norm(shear)
    voronoi_dilatation = (dilatation[:-1] + dilatation[1:])/2
    return dilatation, voronoi_dilatation

@njit(cache=True)
def compare_difference(position, previous_position):
    difference = np.zeros(position.shape[1])
    for i in range(difference.shape[0]):
        difference[i] = np.sqrt(
            (position[0, i]-previous_position[0, i])**2 +
            (position[1, i]-previous_position[1, i])**2 +
            (position[2, i]-previous_position[2, i])**2
        )
    return np.sum(difference)

@njit(cache=True)
def calculate_length(position):
    length = 0
    for i in range(position.shape[1]-1):
        length += np.sqrt(
            (position[0, i+1] - position[0, i])**2 +
            (position[1, i+1] - position[1, i])**2 +
            (position[2, i+1] - position[2, i])**2
        )
    return length

# @njit(cache=True)
# def calculate_potential_energy(strain, curvature, shear_matrix, bend_matrix):
#     dilatation, voronoi_dilatation = calculate_dilatation(strain)
#     energy = 0
#     intrinsic_strain = np.array([0, 0, 1])
#     for i in range(strain.shape[1]):
#         for index_i in range(3):
#             energy += (
#                 shear_matrix[index_i, index_i, i]/dilatation[i]
#                 * (strain[index_i, i]-intrinsic_strain[index_i])**2
#             )
#     for i in range(curvature.shape[1]):
#         for index_i in range(3):
#             energy += (
#                 bend_matrix[index_i, index_i, i]/(voronoi_dilatation[i]**2)
#                 * curvature[index_i, i]**2
#             )
#     return energy/2

# @njit(cache=True)
# def calculate_desired_potential_energy(
#     strain, curvature, shear_matrix, bend_matrix,
#     intrinsic_strain, intrinsic_curvature
# ):
#     dilatation, voronoi_dilatation = calculate_dilatation(strain)
#     energy = 0
#     for i in range(strain.shape[1]):
#         for index_i in range(3):
#             energy += (
#                 shear_matrix[index_i, index_i, i]/dilatation[i]
#                 * (strain[index_i, i]-intrinsic_strain[index_i, i])**2
#             )
#     for i in range(curvature.shape[1]):
#         for index_i in range(3):
#             energy += (
#                 bend_matrix[index_i, index_i, i]/(voronoi_dilatation[i]**2)
#                 * (curvature[index_i, i]-intrinsic_curvature[index_i, i])**2
#             )
#     return energy/2

@njit(cache=True)
def calculate_kinetic_energy(velocity, angular_velocity, mass, inertia):
    energy = 0
    for i in range(velocity.shape[1]):
        for index_i in range(3):
            energy += mass[i]*(velocity[index_i, i]**2)
        for index_i in range(3):
            energy += inertia[index_i, index_i, i] * (angular_velocity[index_i, i]**2)
    return energy/2

@njit(cache=True)
def _vector_function_innder_product(vec_func_1, vec_func_2):
    result = 0
    for i in range(3):
        result += np.trapz(vec_func_1[i]*vec_func_2[i])
    return result

@njit(cache=True)
def sigma_to_shear(sigma):
    shear = np.zeros(sigma.shape)
    for i in range(shear.shape[1]):
        shear[0, i] = sigma[0, i]
        shear[1, i] = sigma[1, i]
        shear[2, i] = sigma[2, i] + 1
    return shear

@njit(cache=True)
def kappa_to_curvature(kappa, voronoi_dilatation):
    curvature = np.zeros(kappa.shape)
    for i in range(curvature.shape[1]):
        curvature[0, i] = kappa[0, i] / voronoi_dilatation[i]
        curvature[1, i] = kappa[1, i] / voronoi_dilatation[i]
        curvature[2, i] = kappa[2, i] / voronoi_dilatation[i]
    return curvature

class CalculateStrainRate(NoForces):
    def __init__(self, sigma_rate, kappa_rate):
        self.sigma_rate = sigma_rate
        self.kappa_rate = kappa_rate
        self.prev_sigma = np.zeros(self.sigma_rate.shape)
        self.prev_kappa = np.zeros(self.kappa_rate.shape)
        self.prev_time = -np.inf

    def apply_torques(self, system, time: float = 0.0):
        # dt = time - self.prev_time
        # self.sigma_rate[:, :] = (system.sigma - self.prev_sigma) / dt
        # self.kappa_rate[:, :] = (system.kappa - self.prev_kappa) / dt

        # self.prev_sigma = system.sigma.copy()
        # self.prev_kappa = system.kappa.copy()
        # self.prev_time = time

        self.calculate_sigma_rate(
            system.rest_lengths, system.dilatation, system.director_collection,
            system.velocity_collection, system.omega_collection,
            system.sigma, self.sigma_rate
        )
        self.calculate_kappa_rate(
            system.rest_voronoi_lengths, system.voronoi_dilatation,
            system.omega_collection, system.kappa, self.kappa_rate
        )
        pass

    @staticmethod
    @njit(cache=True)
    def calculate_sigma_rate(
        rest_lengths, dilatation, director,
        velocity, angular_velocity,
        sigma, sigma_rate
    ):
        sigma_rate[:, :] = _batch_matvec(
            director,
            difference_kernel(velocity)[:, 1:-1]/(rest_lengths*dilatation)
        ) - _batch_cross(angular_velocity, sigma_to_shear(sigma))

    @staticmethod
    @njit(cache=True)
    def calculate_kappa_rate(
        rest_voronoi_lengths, voronoi_dilatation,
        angular_velocity, kappa, kappa_rate
    ):
        kappa_rate[:, :] = (
            difference_kernel(angular_velocity)[:, 1:-1] / (rest_voronoi_lengths*voronoi_dilatation)
             - 2 * _batch_cross(
                quadrature_kernel(angular_velocity)[:, 1:-1],
                kappa
             )
        )



from dataclasses import dataclass
import numpy as np
from numba import njit
from elastica._rotations import _rotate

from elastica.boundary_conditions import ConstraintBase


class BodyBoundaryCondition(ConstraintBase):
    """
    This boundary condition class fixes body orientation

        Attributes
        ----------
        fixed_positions : numpy.ndarray
            2D (dim, 1) array containing data with 'float' type.
        fixed_directors : numpy.ndarray
            3D (dim, dim, 1) array containing data with 'float' type.
    """

    def __init__(self, fixed_position, fixed_director, **kwargs):
        super().__init__(**kwargs)
        self.fixed_position = fixed_position
        self.fixed_director = fixed_director

    def constrain_values(self, rod, time):
        self.compute_contrain_values(
            rod.position_collection,
            self.fixed_position,
            rod.director_collection,
            self.fixed_director,
        )

    def constrain_rates(self, rod, time):
        self.compute_constrain_rates(
            rod.velocity_collection,
            rod.omega_collection,
            rod.acceleration_collection,
            rod.alpha_collection,
        )

    @staticmethod
    @njit(cache=True)
    def compute_contrain_values(
        position_collection, fixed_position, director_collection, fixed_director
    ):
        # Position
        position_collection[2, 0] = fixed_position[2]
        # Director
        director_collection[2, 0, 0] = 0.0
        director_collection[2, 1, 0] = 0.0
        director_collection[2, 2, 0] = 1.0
        for i in range(2):
            length = np.sqrt(
                director_collection[i, 0, 0] ** 2 + director_collection[i, 1, 0] ** 2
            )
            for j in range(2):
                director_collection[i, j, 0] /= length
            director_collection[i, 2, 0] = 0.0

    @staticmethod
    @njit(cache=True)
    def compute_constrain_rates(velocity, omega, acceleration, alpha):
        """
        Compute contrain rates in numba njit decorator
        Parameters
        ----------
        velocity : numpy.ndarray
            2D (dim, blocksize) array containing data with `float` type.
        omega : numpy.ndarray
            2D (dim, blocksize) array containing data with `float` type.
        acceleration : numpy.ndarray
            2D (dim, blocksize) array containing data with `float` type.
        alpha : numpy.ndarray
            2D (dim, blocksize) array containing data with `float` type.

        Returns
        -------

        """
        # Translational
        velocity[2, :] = 0.0
        #acceleration[2, :] = 0.0
        # Rotational
        omega[:2, :] = 0.0
        #alpha[:2, :] = 0.0





import gymnasium as gym

import argparse


def main():
    parser = argparse.ArgumentParser(
        description="Make registered environment and test run."
    )
    parser.add_argument("--env", type=str, default="OctoFlat-v0")
    args = parser.parse_args()

    # env is created, now we can use it:
    env = gym.make(args.env)

    for episode in range(10):
        observation, _ = env.reset()
        for step in range(50):
            action = env.action_space.sample()
            observation, reward, done, truncated, info = env.step(action)
            print(f"{episode=:2} |{step=:2}, {reward=}, {done=}")
            if done:
                break


if __name__ == "__main__":
    main()



"""
Created on Oct. 19, 2020
@author: Heng-Sheng (Hanson) Chang
"""

from gym_softrobot.utils.actuation.algorithms.algorithm import *
from gym_softrobot.utils.actuation.algorithms.forward_backward import *
from gym_softrobot.utils.actuation.algorithms.forward_backward_muscle import *
from gym_softrobot.utils.actuation.algorithms.smoothing_algorithm2 import *



import pytest
import numpy as np

from gym import envs
from gym.spaces import Box
from gym.utils.env_checker import check_env

from tests.envs.spec_list import spec_list

from gym_softrobot import RENDERER_CONFIG
from gym_softrobot.config import RendererType

RENDERER_CONFIG = RendererType.MATPLOTLIB


# This runs a smoketest on each official registered env. We may want
# to try also running environments which are not officially registered
# envs.
@pytest.mark.parametrize("spec", spec_list)
def test_env(spec):
    # Capture warnings
    with pytest.warns(None) as warnings:
        env = spec.make()

    # Test if env adheres to Gym API
    check_env(env, warn=True, skip_render_check=True)

    # Check that dtype is explicitly declared for gym.Box spaces
    for warning_msg in warnings:
        assert "autodetected dtype" not in str(warning_msg.message)

    ob_space = env.observation_space
    act_space = env.action_space
    ob = env.reset()
    assert ob_space.contains(ob), f"Reset observation: {ob!r} not in space"
    if isinstance(ob_space, Box):
        # Only checking dtypes for Box spaces to avoid iterating through tuple entries
        assert (
            ob.dtype == ob_space.dtype
        ), f"Reset observation dtype: {ob.dtype}, expected: {ob_space.dtype}"

    a = act_space.sample()
    observation, reward, done, _info = env.step(a)
    assert ob_space.contains(
        observation
    ), f"Step observation: {observation!r} not in space"
    assert np.isscalar(reward), f"{reward} is not a scalar for {env}"
    assert isinstance(done, bool), f"Expected {done} to be a boolean"
    if isinstance(ob_space, Box):
        assert (
            observation.dtype == ob_space.dtype
        ), f"Step observation dtype: {ob.dtype}, expected: {ob_space.dtype}"

    # FIXME: Test rendering need to install povray on CI. It is disabled for now.
    #for mode in env.metadata.get("render.modes", []):
    #    env.render(mode=mode)

    # Make sure we can render the environment after close.
    #for mode in env.metadata.get("render.modes", []):
    #    env.render(mode=mode)

    env.close()

@pytest.mark.parametrize("spec", spec_list)
def test_reset_info(spec):

    with pytest.warns(None) as warnings:
        env = spec.make()

    ob_space = env.observation_space
    obs = env.reset()
    assert ob_space.contains(obs)
    obs = env.reset(return_info=False)
    assert ob_space.contains(obs)
    obs, info = env.reset(return_info=True)
    assert ob_space.contains(obs)
    assert isinstance(info, dict)
    env.close()

# FIXME: Current version of SB3 1.4.0 uses older gym version, which is causing some issue with 
# newer gym version 0.23.1. This block is commented out until those conflicts are solved.
#@pytest.mark.parametrize("spec", spec_list)
#def test_sb3_check_env(spec):
#    from stable_baselines3.common.env_checker import check_env
#    # Run SB3's check_env function for compatiblity check
#    with pytest.warns(None) as warnings:
#        env = spec.make()
#    # It will check your custom environment and output additional warnings if needed
#    check_env(env)



from gym_softrobot.envs.octopus.flat_env import FlatEnv
from gym_softrobot.envs.octopus.arm_single_env import ArmSingleEnv
from gym_softrobot.envs.octopus.arm_two_env import ArmTwoEnv
from gym_softrobot.envs.octopus.arm_push_env import ArmPushEnv
from gym_softrobot.envs.octopus.arm_push_env import ArmPullWeightEnv
from gym_softrobot.envs.octopus.reach_env import ReachEnv
from gym_softrobot.envs.octopus.crawl_env import CrawlEnv
# from gym_softrobot.envs.octopus.crawlCurvature_env import CrawlCurvatureEnv



"""
Created on Jan. 04, 2021
@author: Heng-Sheng (Hanson) Chang
"""

import numpy as np
from numba import njit

from elastica._linalg import _batch_matvec, _batch_cross
from elastica._calculus import quadrature_kernel, difference_kernel
from elastica.external_forces import inplace_addition, NoForces

@njit(cache=True)
def _material_to_lab(director_collection, vectors):
    blocksize = vectors.shape[1]
    lab_frame_vectors = np.zeros((3, blocksize))
    for n in range(blocksize):
        for i in range(3):
            for j in range(3):
                lab_frame_vectors[i, n] += (
                    director_collection[j, i, n] * vectors[j, n]
                )
    return lab_frame_vectors

@njit(cache=True)
def _lab_to_material(director_collection, vectors):
    return _batch_matvec(director_collection, vectors)

@njit(cache=True)
def _internal_to_external_load(
    director_collection, kappa, tangents,
    rest_lengths, rest_voronoi_lengths,
    dilatation, voronoi_dilatation,
    internal_forces, internal_couples,
    external_forces, external_couples
    ):

    external_forces[:, :] = difference_kernel(
        _material_to_lab(director_collection, internal_forces)
        )

    external_couples[:, :] = (
        difference_kernel(internal_couples) +
        quadrature_kernel(
            _batch_cross(kappa, internal_couples) * rest_voronoi_lengths
        ) +
        _batch_cross(
                _lab_to_material(director_collection, tangents * dilatation),
                internal_forces
            )  * rest_lengths
    )
    return

@njit(cache=True)
def _force_induced_couple(internal_forces, distance, internal_couples):
    internal_couples[:, :] = quadrature_kernel(
        _batch_cross(distance, internal_forces)
    )[:, 1:-1]
    return

class ContinuousActuation(object):
    """
    This class and the classes ingerited from this should contain the parameters
    and functions that describe and calculate the forces/couples generated from
    the actuation.
    """

    def __init__(self, n_elements: int):
        self.internal_forces = np.zeros((3, n_elements))       # material frame
        self.external_forces = np.zeros((3, n_elements+1))     # global frame
        self.internal_couples = np.zeros((3, n_elements-1))    # material frame
        self.external_couples = np.zeros((3, n_elements))      # material frame
    
    def __call__(self, system):
        raise NotImplementedError
    
class ApplyActuation(NoForces):
    def __init__(self, actuation, step_skip: int, callback_params: dict):
        self.current_step = 0
        self.actuation = actuation
        self.every = step_skip
        self.callback_params = callback_params

    def apply_torques(self, system, time: float = 0.0):
        
        self.actuation(system)
        inplace_addition(system.external_forces, self.actuation.external_forces)
        inplace_addition(system.external_torques, self.actuation.external_couples)

        self.make_callback()

    def make_callback(self,):
        if self.current_step % self.every == 0:
            self.callback_func()
        self.current_step += 1

    def callback_func(self):
        self.callback_params['internal_forces'].append(self.actuation.internal_forces.copy())
        self.callback_params['internal_couples'].append(self.actuation.internal_couples.copy())
        self.callback_params['external_forces'].append(self.actuation.external_forces.copy())
        self.callback_params['external_couples'].append(self.actuation.external_couples.copy())



"""
Created on Oct. 19, 2020
@author: Heng-Sheng (Hanson) Chang
"""

import numpy as np

class Constraint(object):
    def __init__(self, constraint_type, constraint_cost_weight):
        self.type = constraint_type
        self.constraint_cost_weight = constraint_cost_weight
        self.cost = np.zeros(1)

class Target(object):
    def __init__(self, target_type, position, director, target_cost_weight):
        self.type = target_type
        self.position = np.array(position)
        self.director = np.array(director)
        self.target_cost_weight = target_cost_weight
        self.cost = np.zeros(1)



__doc__ = """
Module contains elastica interface to create octopus model.

We tried to select the mechanical parameters within the plausible range in soft-matter studies
and simplified model. We encourage users to try out different parameters for properties, but
the behavior of physical simulator (PyElastica) might be very different.

"""

from typing import Optional

import numpy as np

from elastica import *

from gym_softrobot.utils.custom_elastica.joint import FixedJoint2Rigid
from gym_softrobot.utils.custom_elastica.constraint import BodyBoundaryCondition
from gym_softrobot.utils.actuation.actuations.muscles.longitudinal_muscle import (
    LongitudinalMuscle,
)
from gym_softrobot.utils.actuation.actuations.muscles.transverse_muscle import (
    TransverseMuscle,
)

from scipy.spatial.transform import Rotation as Rot

_OCTOPUS_PROPERTIES = {  # default parameters
    # Arm properties
    "youngs_modulus": 1e6,
    "density": 1000.0,
    # "nu": 1e-2,  # Deprecated
    # "poisson_ratio": 0.5,  # Deprecated
    # Head properties
    "body_arm_k": 1e6,
    "body_arm_kt": 1e0,
    "head_radius": 0.04,
    "head_density": 700.0,
    # Friction Properties
    "friction_multiplier": 1.00,
    "friction_symmetry": False,
}

_DEFAULT_SCALE_LENGTH = {
    "base_length": 0.35,
    "base_radius": 0.35 * 0.02,
}


def build_octopus(
    simulator,
    n_arm: int = 8,
    n_elem: int = 11,
    time_step: float = 7e-5,
    override_params: Optional[dict] = None,
):
    """Import default parameters (overridable)"""
    param = _OCTOPUS_PROPERTIES.copy()  # Always copy parameter for safety
    if isinstance(override_params, dict):
        param.update(override_params)
    """ Import default parameters (non-overridable) """
    arm_scale_param = _DEFAULT_SCALE_LENGTH.copy()

    """ Set up an arm """
    L0 = arm_scale_param["base_length"]
    r0 = arm_scale_param["base_radius"]

    rigid_rod_length = r0 * 2
    rigid_rod_radius = param["head_radius"]

    rotation_angle = 360 / n_arm
    angle_list = [rotation_angle * arm_i for arm_i in range(n_arm)]

    shearable_rods = []  # arms
    for arm_i in range(n_arm):
        arm_angle = angle_list[arm_i]
        rot = Rot.from_euler("z", arm_angle, degrees=True)
        arm_pos = rot.apply([rigid_rod_radius, 0.0, 0.0])
        arm_dir = rot.apply([1.0, 0.0, 0.0])
        rod = CosseratRod.straight_rod(
            n_elements=n_elem,
            start=arm_pos,
            direction=arm_dir,
            normal=np.array([0.0, 0.0, 1.0]),
            **arm_scale_param,
            **param,
            # nu_for_torques=damp_coefficient*((radius_mean/radius_base)**4),
        )
        shearable_rods.append(rod)
        simulator.append(rod)

    """ Add head """
    start = np.zeros((3,))
    start[2] = -r0
    direction = np.array([0.0, 0.0, 1.0])
    normal = np.array([0.0, 1.0, 0.0])
    binormal = np.cross(direction, normal)
    base_area = np.pi * rigid_rod_radius**2
    density = param["head_density"]

    rigid_rod = Cylinder(
        start, direction, normal, rigid_rod_length, rigid_rod_radius, density
    )
    simulator.append(rigid_rod)

    """ Constraint body """
    simulator.constrain(rigid_rod).using(
        # Upright rigid rod need restoration force/torque against the floor
        BodyBoundaryCondition,
        constrained_position_idx=(0,),
        constrained_director_idx=(0,),
    )

    """ Set up boundary conditions """
    for arm_i in range(n_arm):
        _k = param["body_arm_k"]
        _kt = param["body_arm_kt"]
        simulator.connect(
            first_rod=rigid_rod,
            second_rod=shearable_rods[arm_i],
            first_connect_idx=-1,
            second_connect_idx=0,
        ).using(
            FixedJoint2Rigid,
            k=_k,
            nu=1e-3,
            kt=_kt,
            angle=angle_list[arm_i],
            radius=rigid_rod_radius,
        )

    """Add gravity forces"""
    _g = -9.81
    gravitational_acc = np.array([0.0, 0.0, _g])
    for arm_i in range(n_arm):
        simulator.add_forcing_to(shearable_rods[arm_i]).using(
            GravityForces, acc_gravity=gravitational_acc
        )
    # simulator.add_forcing_to(rigid_rod).using(
    #             GravityForces, acc_gravity=gravitational_acc
    #         )

    """ Add damping """
    damping_constant = 1e-2
    for arm_i in range(n_arm):
        simulator.dampen(arm_i).using(
            AnalyticalLinearDamper,
            damping_constant=damping_constant,
            time_step=time_step,
        )

    """ Add drag force """
    # dl = L0 / n_elem
    # fluid_factor = 1
    # r_bar = (radius_base + radius_tip) / 2
    # sea_water_dentsity = 1022
    # c_per = 0.41 / sea_water_dentsity / r_bar / dl * fluid_factor
    # c_tan = 0.033 / sea_water_dentsity / np.pi / r_bar / dl * fluid_factor
    #
    # simulator.add_forcing_to(self.shearable_rod).using(
    #     DragForce,
    #     rho_environment=sea_water_dentsity,
    #     c_per=c_per,
    #     c_tan=c_tan,
    #     system=self.shearable_rod,
    #     step_skip=self.step_skip,
    #     callback_params=self.rod_parameters_dict
    # )

    """Add friction forces (always the last thing before finalize)"""
    normal = np.array([0.0, 0.0, 1.0])
    period = 2.0

    origin_plane = np.array([0.0, 0.0, -r0])
    normal_plane = normal
    slip_velocity_tol = 1e-8
    froude = 0.1
    mu = L0 / (period * period * np.abs(_g) * froude)
    if param["friction_symmetry"]:
        kinetic_mu_array = (
            np.array([mu, mu, mu]) * param["friction_multiplier"]
        )  # [forward, backward, sideways]
    else:
        kinetic_mu_array = (
            np.array([mu, 1.5 * mu, 2.0 * mu]) * param["friction_multiplier"]
        )  # [forward, backward, sideways]
    static_mu_array = 2 * kinetic_mu_array
    for arm_i in range(n_arm):
        simulator.add_forcing_to(shearable_rods[arm_i]).using(
            AnisotropicFrictionalPlane,
            k=1e2,
            nu=1e1,
            plane_origin=origin_plane,
            plane_normal=normal_plane,
            slip_velocity_tol=slip_velocity_tol,
            static_mu_array=static_mu_array,
            kinetic_mu_array=kinetic_mu_array,
        )
    """
    mu = L0 / (period * period * np.abs(_g) * froude)
    kinetic_mu_array = np.array([mu, mu, mu])  # [forward, backward, sideways]
    static_mu_array = 2 * kinetic_mu_array
    simulator.add_forcing_to(rigid_rod).using(
        AnisotropicFrictionalPlaneRigidBody,
        k=8e2,
        nu=1e1,
        plane_origin=origin_plane,
        plane_normal=normal_plane,
        slip_velocity_tol=slip_velocity_tol,
        static_mu_array=static_mu_array,
        kinetic_mu_array=kinetic_mu_array,
    )
    """

    return shearable_rods, rigid_rod


def build_arm(
    simulator,
    n_elem: int = 11,
    time_step: float = 7e-5,
    override_params: Optional[dict] = None,
    attach_head: bool = None,  # TODO: To be implemented
    attach_weight: Optional[bool] = None,  # TODO: To be implemented
):
    """Import default parameters (overridable)"""
    param = _OCTOPUS_PROPERTIES.copy()  # Always copy parameter for safety
    if isinstance(override_params, dict):
        param.update(override_params)
    """ Import default parameters (non-overridable) """
    arm_scale_param = _DEFAULT_SCALE_LENGTH.copy()

    """ Set up an arm """
    L0 = arm_scale_param["base_length"]
    r0 = arm_scale_param["base_radius"]

    arm_pos = np.array([0.0, 0.0, 0.0])
    arm_dir = np.array([1.0, 0.0, 0.0])
    normal = np.array([0.0, 0.0, 1.0])
    rod = CosseratRod.straight_rod(
        n_elements=n_elem,
        start=arm_pos,
        direction=arm_dir,
        normal=normal,
        **arm_scale_param,
        **param,
    )
    simulator.append(rod)

    """Add gravity forces"""
    _g = -9.81
    gravitational_acc = np.array([0.0, 0.0, _g])
    simulator.add_forcing_to(rod).using(GravityForces, acc_gravity=gravitational_acc)

    """Add friction forces (always the last thing before finalize)"""
    contact_k = 1e2  # TODO: These need to be global parameter to tune
    contact_nu = 1e1
    period = 2.0
    origin_plane = np.array([0.0, 0.0, -r0])
    slip_velocity_tol = 1e-8
    froude = 0.1
    mu = L0 / (period * period * np.abs(_g) * froude)
    if param["friction_symmetry"]:
        kinetic_mu_array = (
            np.array([mu, mu, mu]) * param["friction_multiplier"]
        )  # [forward, backward, sideways]
    else:
        kinetic_mu_array = (
            np.array([mu, 1.5 * mu, 2.0 * mu]) * param["friction_multiplier"]
        )  # [forward, backward, sideways]
    static_mu_array = 2 * kinetic_mu_array
    simulator.add_forcing_to(rod).using(
        AnisotropicFrictionalPlane,
        k=contact_k,
        nu=contact_nu,
        plane_origin=origin_plane,
        plane_normal=normal,
        slip_velocity_tol=slip_velocity_tol,
        static_mu_array=static_mu_array,
        kinetic_mu_array=kinetic_mu_array,
    )

    damping_constant = 1e-2
    simulator.dampen(rod).using(
        AnalyticalLinearDamper,
        damping_constant=damping_constant,
        time_step=time_step,
    )

    return rod


def create_es_muscle_layers(
    radius_mean,
    radius_base,
):
    muscle_layers = [
        # LongitudinalMuscle(
        #     muscle_radius_ratio=np.stack(
        #         (np.zeros(radius_mean.shape),
        #          2 / 3 * np.ones(radius_mean.shape)),
        #         axis=0),
        #     max_force=1 * (radius_mean / radius_base) ** 2,
        # )
        LongitudinalMuscle(
            muscle_radius_ratio=np.stack(
                (np.zeros(radius_mean.shape), 6 / 9 * np.ones(radius_mean.shape)),
                axis=0,
            ),
            max_force=0.5 * (radius_mean / radius_base) ** 2,
        ),
        LongitudinalMuscle(
            muscle_radius_ratio=np.stack(
                (np.zeros(radius_mean.shape), -6 / 9 * np.ones(radius_mean.shape)),
                axis=0,
            ),
            max_force=0.5 * (radius_mean / radius_base) ** 2,
        ),
        TransverseMuscle(
            muscle_radius_ratio=np.stack(
                (np.zeros(radius_mean.shape), 4 / 9 * np.ones(radius_mean.shape)),
                axis=0,
            ),
            max_force=1.0 * (radius_mean / radius_base) ** 2,
        ),
    ]
    return muscle_layers



from typing import Optional

from gym import core
from gym import spaces
from gym.utils import seeding

from collections import defaultdict
import time

import numpy as np

from elastica import *
from elastica.timestepper import extend_stepper_interface
from elastica._calculus import _isnan_check

from gym_softrobot import RENDERER_CONFIG
from gym_softrobot.config import RendererType
from gym_softrobot.envs.soft_pendulum.build import build_soft_pendulum
from gym_softrobot.utils.custom_elastica.callback_func import (
    RodCallBack,
)
from gym_softrobot.utils.render.post_processing import plot_video
from gym_softrobot.utils.render.base_renderer import (
    BaseRenderer,
    BaseElasticaRendererSession,
)


class BaseSimulator(
    BaseSystemCollection,
    Constraints,
    Connections,
    Forcing,
    Damping,
    CallBacks,
):
    pass


class SoftPendulumEnv(core.Env):
    """
    Description:
    Source:
    Observation:
    Actions:
    Reward:
    Starting State:
    Episode Termination:
    Solved Requirements:
    """

    metadata = {"render.modes": ["rgb_array"]}

    def __init__(
        self,
        final_time=5.0,
        time_step=1.0e-4,
        recording_fps=25,
        n_elems=50,
        config_generate_video=False,
    ):
        # Integrator type

        self.final_time = final_time
        self.time_step = time_step
        self.total_steps = int(self.final_time / self.time_step)
        self.recording_fps = recording_fps
        self.step_skip = int(1.0 / (recording_fps * time_step))

        self.n_elems = n_elems
        self.n_seg = n_elems - 1

        # Spaces
        self.n_action = 1
        action_size = (self.n_action,)
        action_low = np.ones(action_size) * (-22)
        action_high = np.ones(action_size) * (22)
        self.action_space = spaces.Box(
            action_low, action_high, shape=action_size, dtype=np.float32
        )
        self._observation_size = ((4),)  # 2 for target
        self.observation_space = spaces.Box(
            -np.inf, np.inf, shape=self._observation_size, dtype=np.float32
        )

        self.metadata = {}
        self.reward_range = 100.0
        self._prev_action = np.zeros(
            list(self.action_space.shape), dtype=self.action_space.dtype
        )

        # Configurations
        self.config_generate_video = config_generate_video

        # Rendering-related
        self.viewer = None
        self.renderer = None

        # Determinism
        self.seed()

    def seed(self, seed=None):
        # Deprecated in new gym
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        return_info: bool = False,
        options: Optional[dict] = None,
    ):
        super().reset(seed=seed)
        self.simulator = BaseSimulator()

        self.point_force = np.zeros((self.n_action))

        self.shearable_rod = build_soft_pendulum(
            self.simulator,
            self.n_elems,
            self.point_force,
            self.np_random,
            self.time_step,
        )

        # CallBack
        if self.config_generate_video:
            self.rod_parameters_dict = defaultdict(list)
            self.simulator.collect_diagnostics(self.shearable_rod).using(
                RodCallBack,
                step_skip=self.step_skip,
                callback_params=self.rod_parameters_dict,
            )

        """ Finalize the simulator and create time stepper """
        self.StatefulStepper = PositionVerlet()
        self.simulator.finalize()
        self.do_step, self.stages_and_updates = extend_stepper_interface(
            self.StatefulStepper, self.simulator
        )

        self.time = np.float64(0.0)
        self.counter = 0

        # Initial State
        state = self.get_state()

        if return_info:
            return state, {}
        else:
            return state

    def get_state(self):
        # Build state
        rod = self.shearable_rod
        pos_state1 = rod.position_collection[0, 0]
        vel_state1 = rod.velocity_collection[0, 0]
        tangents_mean = np.mean(self.shearable_rod.tangents, axis=1)
        theta = np.arctan(tangents_mean[0] / tangents_mean[1])  # Target angle is 0
        theta = ((theta + np.pi) % (2 * np.pi)) - np.pi
        previous_action = self._prev_action.copy()
        state = np.hstack([pos_state1, vel_state1, previous_action, theta]).astype(
            np.float32
        )
        return state

    def set_action(self, action) -> None:
        # FIXME: why do you set current action to previous action?
        self._prev_action[:] = action
        self.point_force[:] = action
        # action = np.concatenate([[0], action, [0]], axis=-1)
        # action = interp1d(
        #         np.linspace(0,1,self.n_action+2), # added zero on the boundary
        #         action,
        #         kind='cubic',
        #         axis=-1,
        #     )(np.linspace(0,1,self.n_seg))
        # self.shearable_rod.rest_kappa[0, :] = action # Planar curvature

    def step(self, action):
        self.set_action(action)

        """ Post-simulation """

        """ Run the simulation for one step """
        stime = time.perf_counter()
        for _ in range(self.step_skip):
            self.time = self.do_step(
                self.StatefulStepper,
                self.stages_and_updates,
                self.simulator,
                self.time,
                self.time_step,
            )
        etime = time.perf_counter()
        # print(f'{self.counter=}, {etime-stime}sec, {self.time=}')

        """ Done is a boolean to reset the environment before episode is completed """
        terminated = False
        truncated = False
        survive_reward = 0.0
        forward_reward = 0.0
        # FIXME: How to set control penalty
        control_penalty = 0.0  # 0.005 * np.square(rest_kappa.ravel()).mean()
        # Position of the rod cannot be NaN, it is not valid, stop the simulation
        invalid_values_condition = _isnan_check(
            np.concatenate(
                [
                    self.shearable_rod.position_collection,
                    self.shearable_rod.velocity_collection,
                ]
            )
        )

        if invalid_values_condition:
            print(f" Nan detected in, exiting simulation now. {self.time=}")
            terminated = True
            truncated = True
            survive_reward = -50.0
        else:
            distance_to_origin = np.abs(self.shearable_rod.position_collection[0, 0])
            tangents_mean = np.mean(self.shearable_rod.tangents, axis=1)
            theta = np.arctan(tangents_mean[0] / tangents_mean[1])  # Target angle is 0
            distance_to_target_angle = ((theta + np.pi) % (2 * np.pi)) - np.pi
            forward_reward = distance_to_origin * 10 + distance_to_target_angle**2
            # cm_pos = self.shearable_rod.compute_position_center_of_mass()[:2]
            # dist_to_target = np.linalg.norm(cm_pos - self._target, ord=2)
            # forward_reward = (self.prev_dist_to_target - dist_to_target) * 10
            # self.prev_dist_to_target = dist_to_target
            """ Goal """
            # FIXME: How to set survival reward
            # if distance_to_origin < 0.01  and distance_to_target_angle < 0.01:
            #     survive_reward = 100.0
            #     terminated  = True

        """ Time limit """
        timelimit = False
        if self.time > self.final_time:
            timelimit = True
            terminated = True
            truncated = True

        reward = forward_reward - control_penalty + survive_reward
        # reward *= 10 # Reward scaling
        # print(f'{reward=:.3f}: {forward_reward=:.3f}, {control_panelty=:.3f}, {survive_reward=:.3f}')

        """ Return state:
            (1) current simulation time
            (2) current systems
            (3) a flag denotes whether the simulation runs correlectly
        """
        # systems = [self.shearable_rod]
        states = self.get_state()

        # Info
        info = {
            "time": self.time,
            "rod": self.shearable_rod,
            "TimeLimit.truncated": timelimit,
        }

        self.counter += 1

        return states, reward, terminated, truncated, info

    def save_data(self, filename_video, fps):
        if self.config_generate_video:
            filename_video = f"save/{filename_video}"
            plot_video(self.rod_parameters_dict, filename_video, margin=0.2, fps=fps)

    def render(self, mode="human", close=False):
        maxwidth = 800
        aspect_ratio = 3 / 4

        if self.viewer is None:
            from gym_softrobot.utils.render import pyglet_rendering

            self.viewer = pyglet_rendering.SimpleImageViewer(maxwidth=maxwidth)

        if self.renderer is None:
            # Switch renderer depending on configuration
            if RENDERER_CONFIG == RendererType.POVRAY:
                from gym_softrobot.utils.render.povray_renderer import Session
            elif RENDERER_CONFIG == RendererType.MATPLOTLIB:
                from gym_softrobot.utils.render.matplotlib_renderer import Session
            else:
                raise NotImplementedError("Rendering module is not imported properly")
            assert issubclass(Session, BaseRenderer), (
                "Rendering module is not properly subclassed"
            )
            assert issubclass(Session, BaseElasticaRendererSession), (
                "Rendering module is not properly subclassed"
            )
            self.viewer = pyglet_rendering.SimpleImageViewer(maxwidth=maxwidth)
            self.renderer = Session(width=maxwidth, height=int(maxwidth * aspect_ratio))
            self.renderer.add_rod(
                self.shearable_rod
            )  # TODO: maybe need add_rod instead
            self.renderer.add_point(self._target.tolist() + [0], 0.02)

        # POVRAY
        if RENDERER_CONFIG == RendererType.POVRAY:
            state_image = self.renderer.render(
                maxwidth, int(maxwidth * aspect_ratio * 0.7)
            )
            state_image_side = self.renderer.render(
                maxwidth // 2,
                int(maxwidth * aspect_ratio * 0.3),
                camera_param=("location", [0.0, 0.0, -0.5], "look_at", [0.0, 0, 0]),
            )
            state_image_top = self.renderer.render(
                maxwidth // 2,
                int(maxwidth * aspect_ratio * 0.3),
                camera_param=("location", [0.0, 0.3, 0.0], "look_at", [0.0, 0, 0]),
            )

            state_image = np.vstack(
                [state_image, np.hstack([state_image_side, state_image_top])]
            )
        elif RENDERER_CONFIG == RendererType.MATPLOTLIB:
            state_image = self.renderer.render()
        else:
            raise NotImplementedError("Rendering module is not imported properly")

        self.viewer.imshow(state_image)

        return state_image

    def close(self):
        if self.viewer:
            self.viewer.close()
            self.viewer = None
        if self.renderer:
            self.renderer.close()
            self.renderer = None



import gym
import gym_softrobot
from gym import envs

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

def main():
    spec_list = [
        spec for spec in sorted(envs.registry.all(), key=lambda x:x.id)
        if "gym_softrobot" in spec.entry_point
    ]
    for spec in spec_list:
        print(f"{bcolors.BOLD}{spec.entry_point}{bcolors.ENDC}")
        print(spec)

if __name__=="__main__":
    main()



"""
Created on Mar. 01, 2021
@author: Heng-Sheng (Hanson) Chang
"""

import numpy as np
from numba import njit

from elastica._calculus import quadrature_kernel
from elastica.external_forces import inplace_addition

from elastica.external_forces import NoForces

from gym_softrobot.utils.actuation.rod_tools import _lab_to_material, _material_to_lab, average


class DragForce(NoForces):
    def __init__(self, rho_environment,
        c_per, c_tan, system,
        step_skip: int, callback_params: dict
        ):

        self.rho_environment = rho_environment
        self.c_per = c_per
        self.c_tan = c_tan

        self.scale_per = 0.5 * self.rho_environment * c_per
        self.scale_tan = 0.5 * self.rho_environment * c_tan

        self.velocity_material_frame = np.zeros((3, system.n_elems))
        self.drag_force_material_frame = np.zeros((3, system.n_elems))
        self.drag_force = np.zeros((3, system.n_elems+1))

        self.step = 0
        self.every = step_skip
        self.callback_params = callback_params
    
    def apply_torques(self, system, time: float = 0.0):
        Pa = 2 * system.radius * system.lengths
        Sa = Pa * np.pi
        
        self.calculate_drag_force(
            self.scale_per*Pa, self.scale_tan*Sa,
            system.director_collection, system.velocity_collection,
            self.velocity_material_frame,
            self.drag_force_material_frame, self.drag_force,
        )
        
        inplace_addition(system.external_forces, self.drag_force)
        # self.callback()

    def callback(self,):
        if self.step % self.every == 0:
            self.callback_func()
        self.step += 1

    def callback_func(self):
        self.callback_params['velocity_material_frame'].append(
            self.velocity_material_frame.copy()
        )
        self.callback_params['drag_froce_material_frame'].append(
            self.drag_force_material_frame.copy()
        )

    @staticmethod
    @njit(cache=True)
    def calculate_drag_force(
        scale_per, scale_tan,
        director, velocity,
        velocity_material_frame,
        drag_force_material_frame, drag_force
        ):

        velocity_material_frame[:, :] = (
            _lab_to_material(director, average(velocity))
            )
        square_velocity_with_direction = (
            np.abs(velocity_material_frame) * velocity_material_frame
            )
        drag_force_material_frame[:2, :] = (
            - scale_per * square_velocity_with_direction[:2, :]
            )
        drag_force_material_frame[2, :] = (
            - scale_tan * square_velocity_with_direction[2, :]
            )
        drag_force[:, :] = (
            quadrature_kernel(
                _material_to_lab(director, drag_force_material_frame)
                )
            )



"""
Created on Jan. 07, 2021
@author: Heng-Sheng (Hanson) Chang
"""

import numpy as np
from numba import njit

from elastica._linalg import _batch_cross
from elastica._calculus import quadrature_kernel

from gym_softrobot.utils.actuation.actuations.actuation import _internal_to_external_load, _force_induced_couple
from gym_softrobot.utils.actuation.actuations.muscles.muscle import (
    local_strain,
    MuscleForce,
    MuscleFibers,
)
# from actuations.muscles.activations_function import (
#     set_guassian_activations,
#     GuassianFunction,
#     UniformFunction
# )

# from actuations.muscles.weighted_function import weighted_load
from gym_softrobot.utils.actuation.rod_tools import sigma_to_shear, kappa_to_curvature

class LongitudinalMuscle(MuscleForce):
    def __init__(
        self,
        muscle_radius_ratio,
        max_force,
        strain_weighted_function=None,
        strain_rate_weighted=None
    ):
        MuscleForce.__init__(self, muscle_radius_ratio.shape[1])
        self.muscle_radius_ratio = np.zeros((3, self.n_elements))
        self.muscle_radius_ratio[:2, :] = muscle_radius_ratio.copy()
        self.max_force = max_force.copy()
        self.strain_weighted_function = strain_weighted_function
        self.strain_rate_weighted = False
        if strain_rate_weighted is not None:
            self.strain_rate_weighted = True
            self.strain_rate_weighted_function = strain_rate_weighted['function']
            self.sigma_rate = strain_rate_weighted['sigma_rate']
            self.kappa_rate = strain_rate_weighted['kappa_rate']
        self.count = 0


    def __call__(self, system, count_flag=True):

        magnitude_for_force = self.get_activation() * self.max_force

        weight = self.calculate_force_length_velocity_weight(system)

        # weighted_load(
        #     load=magnitude_for_force,
        #     weight=weight
        # )

        longitudinal_muscle_function(
            magnitude_for_force, self.muscle_radius_ratio*system.radius,
            system.director_collection, system.kappa, system.tangents,
            system.rest_lengths, system.rest_voronoi_lengths,
            system.dilatation, system.voronoi_dilatation,
            self.internal_forces, self.internal_couples,
            self.external_forces, self.external_couples
        )
        if count_flag:
            self.count += 1

    def calculate_force_length_velocity_weight(self, system):
        weight = np.ones(self.activation.shape)

        if self.strain_weighted_function is not None:
            weight[:] = self.strain_weighted_function(
                stretch_mag=self.longitudinal_muscle_stretch_mag(
                    off_center_displacement=self.muscle_radius_ratio*system.radius,
                    strain=sigma_to_shear(system.sigma),
                    curvature=kappa_to_curvature(system.kappa, system.voronoi_dilatation)
                )
            )
        
        # if self.strain_rate_weighted:
        #     stretch_rate = self.longitudinal_muscle_stretch_rate_mag(
        #         off_center_displacement=self.muscle_radius_ratio*system.radius,
        #         sigma_rate=self.sigma_rate,
        #         curvature_rate=kappa_to_curvature(self.kappa_rate, system.voronoi_dilatation)
        #     )
        #     # if (self.count % 1000) == 0:
        #     #     print(np.max(weight))
        #     weight[:] *= self.strain_rate_weighted_function(stretch_rate)

        return weight

    @staticmethod
    @njit(cache=True)
    def longitudinal_muscle_stretch_mag(off_center_displacement, strain, curvature):
        nu = local_strain(
            off_center_displacement, strain, curvature
        )
        blocksize = nu.shape[1]
        stretch_mag = np.ones(blocksize)
        for i in range(blocksize):
            stretch_mag[i] = np.linalg.norm(nu[:, i])
        return stretch_mag

    @staticmethod
    @njit(cache=True)
    def longitudinal_muscle_stretch_rate_mag(off_center_displacement, sigma_rate, curvature_rate):
        nu_rate = local_strain(
            off_center_displacement, sigma_rate, curvature_rate
        )
        blocksize = nu_rate.shape[1]
        stretch_mag = np.ones(blocksize)
        for i in range(blocksize):
            stretch_mag[i] = nu_rate[2, i]
        return stretch_mag

    # @staticmethod
    # @njit(cache=True)
    # def longitudinal_muscle_function(
    #     magnitude_for_force, r_m,
    #     director_collection, kappa, tangents,
    #     rest_lengths, rest_voronoi_lengths,
    #     dilatation, voronoi_dilatation,
    #     internal_forces, internal_couples,
    #     external_forces, external_couples
    # ):

    #     internal_forces[2, :] = magnitude_for_force.copy()
    #     internal_couples[:, :] = quadrature_kernel(
    #         _batch_cross(r_m, internal_forces)
    #     )[:, 1:-1]

    #     _internal_to_external_load(
    #         director_collection, kappa, tangents,
    #         rest_lengths, rest_voronoi_lengths,
    #         dilatation, voronoi_dilatation,
    #         internal_forces, internal_couples,
    #         external_forces, external_couples
    #     )

@njit(cache=True)
def longitudinal_muscle_function(
    magnitude_for_force, r_m,
    director_collection, kappa, tangents,
    rest_lengths, rest_voronoi_lengths,
    dilatation, voronoi_dilatation,
    internal_forces, internal_couples,
    external_forces, external_couples
):

    internal_forces[2, :] = magnitude_for_force.copy()
    _force_induced_couple(internal_forces, r_m, internal_couples)
    # internal_couples[:, :] = quadrature_kernel(
    #     _batch_cross(r_m, internal_forces)
    # )[:, 1:-1]

    _internal_to_external_load(
        director_collection, kappa, tangents,
        rest_lengths, rest_voronoi_lengths,
        dilatation, voronoi_dilatation,
        internal_forces, internal_couples,
        external_forces, external_couples
    )


# class GuassianLongitudinalMuscle(LongitudinalMuscle, GuassianFunction):
#     def __init__(self, muscle_radius_ratio, max_force, variance, **kwargs):
#         LongitudinalMuscle.__init__(self, muscle_radius_ratio, max_force, **kwargs)
#         GuassianFunction.__init__(self, variance)

#     def set_controls(self, controls, positions):
#         self.set_activations(
#             controls, positions, self.s, self.variance, self.activations
#         )

# class UniformLongitudinalMuscle(LongitudinalMuscle, UniformFunction):
#     def __init__(self, muscle_radius_ratio, max_force, width):
#         LongitudinalMuscle.__init__(self, muscle_radius_ratio, max_force)
#         UniformFunction.__init__(self, width)

#     def set_controls(self, controls, positions):
#         self.set_activations(
#             controls, positions, self.s, self.width, self.activations
#         )

# class GuassianLongitudinalMuscleFibers(GuassianLongitudinalMuscle, MuscleFibers):
#     def __init__(self, muscle_radius_ratio, max_force, variance, positions, **kwargs):
#         GuassianLongitudinalMuscle.__init__(self, muscle_radius_ratio, max_force, variance, **kwargs)
#         self.positions = np.array(positions)
#         MuscleFibers.__init__(self, self.n_elements, self.positions.shape[0])

#     def __call__(self, system):
#         if self.G_flag:
#             self.internal_forces[:, :] *= 0
#             self.internal_couples[:, :] *= 0
#             self.external_forces[:, :] *= 0
#             self.external_couples[:, :] *= 0

#             self.muscle_fibers_function(
#                 self.controls,
#                 self.G_internal_forces, self.G_internal_couples,
#                 self.G_external_forces, self.G_external_couples,
#                 self.internal_forces, self.internal_couples,
#                 self.external_forces, self.external_couples
#             )

#             # self.G_activations[:, :] *= 0
#             self.G_internal_forces[:, :, :] *= 0
#             self.G_internal_couples[:, :, :] *= 0
#             self.G_external_forces[:, :, :] *= 0
#             self.G_external_couples[:, :, :] *= 0
#             self.G_flag = False

#         else:
#             GuassianLongitudinalMuscle.__call__(self, system)

#     def set_controls(self, controls):
#         self.controls[:] = np.array(controls)
#         GuassianLongitudinalMuscle.set_controls(self, self.controls, self.positions)
    
#     def calculate_G(self, system):
#         # for i in range(self.positions.shape[0]):
#         #     self.set_activations(
#         #         np.ones(1), self.positions[i:i+1], self.s, self.variance, self.activations
#         #     )
#         #     self(system)
#         #     # self.G_activations[i, :] = self.activations.copy()
#         #     self.G_internal_forces[i, :, :] = self.internal_forces.copy()
#         #     self.G_internal_couples[i, :, :] = self.internal_couples.copy()
#         #     self.G_external_forces[i, :, :] = self.external_forces.copy()
#         #     self.G_external_couples[i, :, :] = self.external_couples.copy()

#         weight = self.calculate_force_length_velocity_weight(system)

#         self.calculate_G_numba(
#             self.positions, self.s, self.variance, self.activations,
#             self.max_force, self.muscle_radius_ratio*system.radius, weight,
#             system.director_collection, system.kappa, system.tangents,
#             system.rest_lengths, system.rest_voronoi_lengths,
#             system.dilatation, system.voronoi_dilatation,
#             self.internal_forces, self.internal_couples,
#             self.external_forces, self.external_couples,
#             self.G_internal_forces, self.G_internal_couples,
#             self.G_external_forces, self.G_external_couples
#         )
        
#         self.G_flag = True
    
#     @staticmethod
#     @njit(cache=True)
#     def calculate_G_numba(
#         positions, s, variance, activations,
#         max_force, off_center_displacement, weight,
#         director_collection, kappa, tangents,
#         rest_lengths, rest_voronoi_lengths,
#         dilatation, voronoi_dilatation,
#         internal_forces, internal_couples,
#         external_forces, external_couples,
#         G_internal_forces, G_internal_couples,
#         G_external_forces, G_external_couples,
#     ):
#         blocksize = positions.shape[0]
#         for i in range(blocksize):
#             set_guassian_activations(
#                 np.ones(1), positions[i: i+1], s, variance, activations
#             )
#             magnitude_for_force = activations * max_force * weight
#             longitudinal_muscle_function(
#                 magnitude_for_force, off_center_displacement,
#                 director_collection, kappa, tangents,
#                 rest_lengths, rest_voronoi_lengths,
#                 dilatation, voronoi_dilatation,
#                 internal_forces, internal_couples,
#                 external_forces, external_couples
#             )
#             G_internal_forces[i, :, :] = internal_forces.copy()
#             G_internal_couples[i, :, :] = internal_couples.copy()
#             G_external_forces[i, :, :] = external_forces.copy()
#             G_external_couples[i, :, :] = external_couples.copy()
#         pass

# class UniformLongitudinalMuscleFibers(UniformLongitudinalMuscle, MuscleFibers):
#     def __init__(self, muscle_radius_ratio, max_force, width, positions):
#         UniformLongitudinalMuscle.__init__(self, muscle_radius_ratio, max_force, width)
#         self.positions = np.array(positions)
#         MuscleFibers.__init__(self, self.n_elements, self.positions.shape[0])

#     def set_controls(self, controls):
#         self.controls[:] = np.array(controls)
#         UniformLongitudinalMuscle.set_controls(self, self.controls, self.positions)
